from typing import Tuple

import numpy as np
import pytest

from helax.numpy import amplitudes, wavefunctions
from helax.numpy.phase_space import PhaseSpace
from helax.numpy.utils import abs2, kallen_lambda
from helax.numpy.wavefunctions import DiracWf, ScalarWf
from helax.vertices import VertexFFS

MUON_MASS = 1.056584e-01
G_FERMI = 1.166379e-05
EL = 3.028620e-01
SW = 4.808530e-01
CW = np.sqrt(1.0 - SW**2)
MASS_W = 8.038500e01
WIDTH_W = 2.085000e00
MASS_B = 4.18
MASS_T = 172.9
MASS_MU = 105.6583745e-3
MASS_E = 0.5109989461e-3
MASS_Z = 91.1876

MASS_H = 125.00
WIDTH_H = 0.00374
VEV_H = 246.21965

SEED = 1234


def spinor_u(momenta, mass: float) -> Tuple[DiracWf, DiracWf]:
    return (
        wavefunctions.spinor_u(momenta, mass, -1),
        wavefunctions.spinor_u(momenta, mass, 1),
    )


def spinor_v(momenta, mass: float) -> Tuple[DiracWf, DiracWf]:
    return (
        wavefunctions.spinor_v(momenta, mass, -1),
        wavefunctions.spinor_v(momenta, mass, 1),
    )


def spinor_ubar(momenta, mass: float) -> Tuple[DiracWf, DiracWf]:
    return (
        wavefunctions.spinor_ubar(momenta, mass, -1),
        wavefunctions.spinor_ubar(momenta, mass, 1),
    )


def spinor_vbar(momenta, mass: float) -> Tuple[DiracWf, DiracWf]:
    return (
        wavefunctions.spinor_vbar(momenta, mass, -1),
        wavefunctions.spinor_vbar(momenta, mass, 1),
    )


def scalar_wf(momenta, out: bool) -> ScalarWf:
    return wavefunctions.scalar_wf(momenta, out)


def current_ff_to_s(vertex, mass, width, psi_out: DiracWf, psi_in: DiracWf):
    return amplitudes.current_ff_to_s(vertex, mass, width, psi_out, psi_in)


def amplitude_ffs(vertex, psi_out: DiracWf, psi_in: DiracWf, phi: ScalarWf):
    return amplitudes.amplitude_ffs(vertex, psi_out, psi_in, phi)


def wavefunction_iter(*wavefuncs):
    idx_list = []
    for wf in wavefuncs:
        if isinstance(wf, wavefunctions.ScalarWf):
            idx_list.append([0])
        elif isinstance(wf, wavefunctions.DiracWf):
            idx_list.append([0, 1])
        elif isinstance(wf, wavefunctions.VectorWf):
            idx_list.append([0, 1, 2])
        else:
            raise ValueError(f"Invalid wf {type(wf)}")

    idxs = np.array(np.meshgrid(*idx_list)).T.reshape(-1, len(idx_list))
    for idx in idxs:
        yield tuple([wf[i] for wf, i in zip(wavefuncs, idx)])


# ============================================================================
# ---- H -> f + f ------------------------------------------------------------
# ============================================================================


def widths_h_to_f_f(mf, ncf):
    yf = mf / VEV_H
    vertex = VertexFFS(
        left=yf / np.sqrt(2),
        right=yf / np.sqrt(2),
    )

    def msqrd_h_to_f_f(momenta):
        p1 = momenta[:, 0]
        p2 = momenta[:, 1]
        ph = np.sum(momenta, axis=1)

        wf1s = spinor_ubar(p1, mf)
        wf2s = spinor_v(p2, mf)
        h_wf = scalar_wf(ph, False)

        def msqrd(i1, i2):
            return abs2(amplitude_ffs(vertex, wf1s[i2], wf2s[i1], h_wf))

        idxs = np.array(np.meshgrid([0, 1], [0, 1])).T.reshape(-1, 2)
        res = sum(msqrd(i1, i2) for (i1, i2) in idxs)

        return ncf * res

    mu = mf / MASS_H
    analytic = ncf * yf**2 * MASS_H * (1 - 4 * mu**2) ** 1.5 / (16 * np.pi)

    phase_space = PhaseSpace(MASS_H, [mf, mf], msqrd=msqrd_h_to_f_f)
    width = phase_space.decay_width(1000, seed=SEED)[0]

    return width, analytic


def test_width_h_to_b_b():
    width, analytic = widths_h_to_f_f(mf=MASS_B, ncf=3.0)
    assert float(width) == pytest.approx(analytic, rel=1e-3, abs=0.0)


# ============================================================================
# ---- chi + chi -> S -> f + f -----------------------------------------------
# ============================================================================


def test_dark_matter_annihilation_scalar_mediator():
    gsxx = 1.0
    gsff = 1.0
    mx = 10.0
    mf = 1.0
    cme = 3 * mx
    mass_s = 1.0
    width_s = 1.0

    vxx = VertexFFS(left=gsxx, right=gsxx)
    vff = VertexFFS(left=gsff, right=gsff)

    ex = cme / 2.0
    px = np.sqrt(ex**2 - mx**2)
    p1 = np.expand_dims(np.array([ex, 0.0, 0.0, px]), -1)
    p2 = np.expand_dims(np.array([ex, 0.0, 0.0, -px]), -1)

    wf1s = spinor_u(p1, mx)
    wf2s = spinor_vbar(p2, mx)

    def msqrd_x_x_to_f_f(momenta):
        p3 = momenta[:, 0]
        p4 = momenta[:, 1]

        wf3s = spinor_ubar(p3, mf)
        wf4s = spinor_v(p4, mf)

        def msqrd(i1, i2, i3, i4):
            swf = current_ff_to_s(vxx, mass_s, width_s, wf2s[i2], wf1s[i1])
            return abs2(amplitude_ffs(vff, wf3s[i3], wf4s[i4], swf))

        idxs = np.array(np.meshgrid([0, 1], [0, 1], [0, 1], [0, 1])).T.reshape(-1, 4)
        res = sum(msqrd(i1, i2, i3, i4) for (i1, i2, i3, i4) in idxs)

        return res / 4.0

    phase_space = PhaseSpace(cme, [mf, mf], msqrd=msqrd_x_x_to_f_f)
    cs = phase_space.cross_section(mx, mx, 10_000, seed=SEED)[0]

    s = cme**2
    analytic = (
        gsff**2
        * gsxx**2
        * (s - 4 * mf**2)
        * np.sqrt((s - 4 * mf**2) * (s - 4 * mx**2))
    ) / (16 * np.pi * s * ((s - mass_s**2) ** 2 + (mass_s * width_s) ** 2))

    assert float(cs) == pytest.approx(analytic, rel=1e-2, abs=0.0)


# ============================================================================
# ---- Ni  -> Nj + phi -------------------------------------------------------
# ============================================================================


def widths_ni_to_nj_phi(mi, mj, mphi, y):

    vertex = VertexFFS(left=y, right=y)

    def msqrd_ni_to_nj_phi(momenta):
        pj = momenta[:, 0]
        pphi = momenta[:, 1]
        pi = np.sum(momenta, axis=1)

        wfi_u = spinor_u(pi, mi)
        wfj_ubar = spinor_ubar(pj, mj)

        wfphi = scalar_wf(pphi, mphi)

        def msqrd(ii, ij):
            amp = amplitude_ffs(vertex, wfj_ubar[ij], wfi_u[ii], wfphi)
            return abs2(amp)

        idxs = np.array(np.meshgrid([0, 1], [0, 1])).T.reshape(-1, 2)
        res = sum(msqrd(ii, ij) for (ii, ij) in idxs)

        return res / 2.0

    rj = mj**2 / mi**2
    rphi = mphi**2 / mi**2

    analytic = (
        mi
        / (16 * np.pi)
        * np.sqrt(kallen_lambda(1, rphi, rj))
        * (np.abs(y) ** 2 * (1 + rj - rphi) + 2 * np.sqrt(rj) * np.real(y**2))
    )

    phase_space = PhaseSpace(mi, [mj, mphi], msqrd=msqrd_ni_to_nj_phi)
    width = phase_space.decay_width(10_000, seed=SEED)[0]

    return width, analytic


def test_width_ni_to_nj_phi():
    mi = 100.0
    mj = 25.0
    mphi = 3.0
    y = 1.0

    width, analytic = widths_ni_to_nj_phi(mi=mi, mj=mj, mphi=mphi, y=y)
    assert float(width / analytic) == pytest.approx(1, rel=1e-3, abs=0.0)



"""Module for computing currents and amplitudes for two fermions and a vector.
"""


# Notes on signs

# All momentum are considered incoming, except for fermions. For fermions, the
# momentum always points along the direction of the fermion flow.

# [v,s](p) + [v,s](q) -> [v,s](k)
# -----------------------
# k = p + q

#     p ->     k=p+q ->
#     -------x------
#           /
#     -----
#     q ->


# f(p) + [v,s](q) -> f(k)
# -----------------------
# k = p + q

#     p ->     ->  k=p+q
#     --->---x--->---
#           /
#     -----
#     q ->


# fbar(p) + [v,s](q) -> fbar(k)
# -----------------------------
# k = p - q

#     p <-     <- k=p-q
#     ---<---x---<---
#           /
#     -----
#     q ->


# f(p) + fbar(q) -> [v,s](k)
# --------------------------
# k = p - q

#     p ->     -> k=p-q
#     --->---x------
#           /
#     --<--
#     q <-


# w(p) + w(q) -> s(k)
# -------------------
# k = p + q

#     p ->     -> k=p+q
#     --->---x------
#           /
#     -->--
#     q ->


# w^+(p) + w^+(q) -> s(k)
# ---------------------------
# k = -p - q

#     p <-     -> k=-p-q
#     ---<---x------
#           /
#     --<--
#     q <-


import numpy as np

from helax.numpy.wavefunctions import DiracWf, VectorWf
from helax.vertices import VertexFFV

from .propagators import attach_dirac, attach_vector

IM = 1.0j

# =============================================================================
# ---- Currents ---------------------------------------------------------------
# =============================================================================


def current_fv_to_f(
    vertex: VertexFFV, mass: float, width: float, psi: DiracWf, polvec: VectorWf
) -> DiracWf:
    """Fuse a fermion and vector into and off-shell fermion.

    The off-shell fermion will have the same fermion flow as `psi`.

    Parameters
    ----------
    vertex : VertexFFV
        Feynman rule for the F-F-V vertex.
    mass : float
        Mass of the produced fermion.
    width : float
        Width of the produced fermion.
    psi : DiracWf
        Fermion to fuse with vector.
    polvec : VectorWf
        Vector to fuse with fermion.

    Returns
    -------
    chi: DiracWf
        Off-shell generated fermion.
    """
    eps = polvec.wavefunction
    f = psi.wavefunction
    vl = vertex.left
    vr = vertex.right

    momentum = psi.momentum * psi.direction * polvec.momentum

    if psi.direction == -1:
        wavefunction = np.array(
            [
                vl * (f[3] * (eps[1] + IM * eps[2]) + f[2] * (eps[0] + eps[3])),
                vl * (f[2] * (eps[1] - IM * eps[2]) + f[3] * (eps[0] - eps[3])),
                vr * (-f[1] * (eps[1] + IM * eps[2]) + f[0] * (eps[0] - eps[3])),
                vr * (-f[0] * (eps[1] - IM * eps[2]) + f[1] * (eps[0] + eps[3])),
            ]
        )
    else:
        wavefunction = np.array(
            [
                vr * (-f[3] * (eps[1] - IM * eps[2]) + f[2] * (eps[0] - eps[3])),
                vr * (-f[2] * (eps[1] + IM * eps[2]) + f[3] * (eps[0] + eps[3])),
                vl * (f[1] * (eps[1] - IM * eps[2]) + f[0] * (eps[0] + eps[3])),
                vl * (f[0] * (eps[1] + IM * eps[2]) + f[1] * (eps[0] - eps[3])),
            ]
        )

    psi = DiracWf(wavefunction=wavefunction, momentum=momentum, direction=psi.direction)
    return attach_dirac(psi, mass, width)


def current_ff_to_v(
    vertex: VertexFFV, mass: float, width: float, psi_out: DiracWf, psi_in: DiracWf
) -> VectorWf:
    """Fuse two fermions into a vector boson.

    Parameters
    ----------
    vertex : VertexFFV
        Feynman rule for the F-F-V vertex.
    mass : float
        Mass of the resulting vector.
    width : float
        Width of the resulting vector.
    psi_out : DiracWf
        Flow-out fermion.
    psi_in : DiracWf
        Flow-in fermion.

    Returns
    -------
    eps: VectorWf
        Resulting vector boson wavefuction.
    """
    wf_in = psi_in.wavefunction
    wf_out = psi_out.wavefunction
    vl = vertex.left
    vr = vertex.right

    momentum = psi_in.momentum - psi_out.momentum

    wavefunction = np.array(
        [
            vr * (wf_in[2] * wf_out[0] + wf_in[3] * wf_out[1])
            + vl * (wf_in[0] * wf_out[2] + wf_in[1] * wf_out[3]),
            vr * (wf_in[3] * wf_out[0] + wf_in[2] * wf_out[1])
            - vl * (wf_in[1] * wf_out[2] + wf_in[0] * wf_out[3]),
            IM * vr * (wf_in[2] * wf_out[1] - wf_in[3] * wf_out[0])
            + IM * vl * (wf_in[1] * wf_out[2] - wf_in[0] * wf_out[3]),
            vr * (wf_in[2] * wf_out[0] - wf_in[3] * wf_out[1])
            + vl * (wf_in[1] * wf_out[3] - wf_in[0] * wf_out[2]),
        ]
    )

    return attach_vector(
        VectorWf(wavefunction=wavefunction, momentum=momentum, direction=1), mass, width
    )


# =============================================================================
# ---- Amplitudes -------------------------------------------------------------
# =============================================================================


def amplitude_ffv(
    vertex: VertexFFV, psi_out: DiracWf, psi_in: DiracWf, polvec: VectorWf
):
    """Compute the scattering amplitude for a Fermion-Fermion-Vector vertex.

    Parameters
    ----------
    vertex : VertexFFV
        Feynman rule for the F-F-V vertex.
    psi_out : DiracWf
        Flow-out fermion wavefunction.
    psi_in : DiracWf
        Flow-in fermion wavefunction.
    polvec : VectorWf
        Vector wavefunction.

    Returns
    -------
    amp: complex
        Scattering amplitude.
    """
    assert psi_out.direction == -1, "`psi_out` must be have flow out."
    assert psi_in.direction == 1, "`psi_in` must be have flow in."

    wf_in = psi_in.wavefunction
    wf_out = psi_out.wavefunction
    eps = polvec.wavefunction
    vl = vertex.left
    vr = vertex.right
    im = 1.0j

    eps0p3 = eps[0] + eps[3]
    eps0m3 = eps[0] - eps[3]

    eps1p2 = eps[1] + im * eps[2]
    eps1m2 = eps[1] - im * eps[2]

    return vr * (
        wf_in[2] * (-wf_out[1] * eps1p2 + wf_out[0] * eps0m3)
        + wf_in[3] * (-wf_out[0] * eps1m2 + wf_out[1] * eps0p3)
    ) + vl * (
        wf_in[1] * (wf_out[2] * eps1m2 + wf_out[3] * eps0m3)
        + wf_in[0] * (wf_out[3] * eps1p2 + wf_out[2] * eps0p3)
    )



import pathlib
from typing import Callable

import jax.numpy as jnp
import numpy as np
from pytest import approx

from helax.jax.wavefunctions import (DiracWf, spinor_u, spinor_ubar, spinor_v,
                                     spinor_vbar, vector_wf)

DWF_DATA = np.load(
    pathlib.Path(__file__).parent.joinpath("testdata").joinpath("spinor_data.npz")
)
VWF_DATA = np.load(
    pathlib.Path(__file__).parent.joinpath("testdata").joinpath("vector_wf_data.npz")
)


def run_spinor_tests(
    fn: Callable[[jnp.ndarray, float, int], DiracWf], ty: str, massive: bool
):
    assert ty in ["u", "v", "ubar", "vbar"], "Invalid string passed to test runner."
    if massive:
        prefix = ty + "_massive_"
        mass = 3.0
    else:
        prefix = ty + "_massless_"
        mass = 0.0

    momenta = jnp.array(DWF_DATA[prefix + "momenta"])
    spin_up: np.ndarray = DWF_DATA[prefix + "up"]
    spin_down: np.ndarray = DWF_DATA[prefix + "down"]

    helax_spin_up: np.ndarray = np.transpose(fn(momenta.T, mass, 1).wavefunction)
    helax_spin_down = np.transpose(fn(momenta.T, mass, -1).wavefunction)

    for tu, td, hu, hd in zip(spin_up, spin_down, helax_spin_up, helax_spin_down):
        for i in range(4):
            assert np.real(hu[i]) == approx(np.real(tu[i]), rel=1e-4, abs=0.0)
            assert np.real(hd[i]) == approx(np.real(td[i]), rel=1e-4, abs=0.0)
            assert np.imag(hu[i]) == approx(np.imag(tu[i]), rel=1e-4, abs=0.0)
            assert np.imag(hd[i]) == approx(np.imag(td[i]), rel=1e-4, abs=0.0)


def test_spinor_u_massive():
    run_spinor_tests(spinor_u, "u", massive=True)


def test_spinor_v_massive():
    run_spinor_tests(spinor_v, "v", massive=True)


def test_spinor_ubar_massive():
    run_spinor_tests(spinor_ubar, "ubar", massive=True)


def test_spinor_vbar_massive():
    run_spinor_tests(spinor_vbar, "vbar", massive=True)


def test_spinor_u_massless():
    run_spinor_tests(spinor_u, "u", massive=False)


def test_spinor_v_massless():
    run_spinor_tests(spinor_v, "v", massive=False)


def test_spinor_ubar_massless():
    run_spinor_tests(spinor_ubar, "ubar", massive=False)


def test_spinor_vbar_massless():
    run_spinor_tests(spinor_vbar, "vbar", massive=False)


def test_vector_wf_massless():
    momenta = jnp.array(VWF_DATA["massless_momenta"])
    spin_up: np.ndarray = VWF_DATA["massless_up"]
    spin_down: np.ndarray = VWF_DATA["massless_down"]

    helax_spin_up: np.ndarray = np.transpose(
        vector_wf(momenta.T, 0.0, 1, False).wavefunction
    )
    helax_spin_down = np.transpose(vector_wf(momenta.T, 0.0, -1, False).wavefunction)

    for t, h in zip(spin_up, helax_spin_up):
        for i in range(4):
            assert np.real(h[i]) == approx(np.real(t[i]), rel=1e-4, abs=0.0)
            assert np.imag(h[i]) == approx(np.imag(t[i]), rel=1e-4, abs=0.0)

    for t, h in zip(spin_down, helax_spin_down):
        for i in range(4):
            assert np.real(h[i]) == approx(np.real(t[i]), rel=1e-4, abs=0.0)
            assert np.imag(h[i]) == approx(np.imag(t[i]), rel=1e-4, abs=0.0)


def test_vector_wf_massive():
    momenta = jnp.array(VWF_DATA["massive_momenta"])
    spin_up: np.ndarray = VWF_DATA["massive_up"]
    spin_zero: np.ndarray = VWF_DATA["massive_zero"]
    spin_down: np.ndarray = VWF_DATA["massive_down"]
    mass = 1.0

    helax_spin_up: np.ndarray = np.transpose(
        vector_wf(momenta.T, mass, 1, False).wavefunction
    )
    helax_spin_zero: np.ndarray = np.transpose(
        vector_wf(momenta.T, mass, 0, False).wavefunction
    )
    helax_spin_down = np.transpose(vector_wf(momenta.T, mass, -1, False).wavefunction)

    for t, h in zip(spin_up, helax_spin_up):
        for i in range(4):
            assert np.real(h[i]) == approx(np.real(t[i]), rel=1e-4, abs=0.0)
            assert np.imag(h[i]) == approx(np.imag(t[i]), rel=1e-4, abs=0.0)

    for t, h in zip(spin_zero, helax_spin_zero):
        for i in range(4):
            assert np.real(h[i]) == approx(np.real(t[i]), rel=1e-4, abs=0.0)
            assert np.imag(h[i]) == approx(np.imag(t[i]), rel=1e-4, abs=0.0)

    for t, h in zip(spin_down, helax_spin_down):
        for i in range(4):
            assert np.real(h[i]) == approx(np.real(t[i]), rel=1e-4, abs=0.0)
            assert np.imag(h[i]) == approx(np.imag(t[i]), rel=1e-4, abs=0.0)



import functools
from typing import NamedTuple, Union

import chex
import jax
import jax.numpy as jnp


class DiracWf(NamedTuple):
    wavefunction: chex.Array
    momentum: chex.Array
    direction: int


class VectorWf(NamedTuple):
    wavefunction: chex.Array
    momentum: chex.Array
    direction: int


class ScalarWf(NamedTuple):
    wavefunction: chex.Array
    momentum: chex.Array
    direction: int


# =============================================================================
# ---- Dirac Wavefunction -----------------------------------------------------
# =============================================================================


def _check_spin_dirac(s: int):
    assert s == 1 or s == -1, "Spin must be 1 or -1."


def _chi(p: chex.Array, s: int):
    """
    Compute the two-component weyl spinor.

    Parameters
    ----------
    p: array
        Array containing the 4-momentum of the wavefunction.
    s: int
        Spin of the wavefunction. Must be 1 or -1.
    """
    eps = jnp.finfo(p.dtype).eps

    px = p[1]
    py = p[2]
    pz = p[3]

    pm = jnp.linalg.norm(p[1:], axis=0)

    def _chi_aligned(s: int, *_):
        # Compute the two-component weyl spinor in the case where pz = -pm.
        # return jnp.array([(s - 1.0) / 2.0 + 0j, (s + 1.0) / 2.0 + 0j])
        return jnp.array([0.0j, s + 0.0j])

    def _chi_unaligned(s, px, py, pz, pm):
        # Compute the two-component weyl spinor in the case where pz != -pm.
        den = jnp.sqrt(2 * pm * (pm + pz))

        return jnp.array(
            [
                (pm + pz) / den,
                (s * px + py * 1j) / den,
            ]
        )

    x = jax.lax.cond(pm + pz < eps, _chi_aligned, _chi_unaligned, s, px, py, pz, pm)

    return jax.lax.switch(
        s + 1,
        [
            lambda: jnp.array([x[1], x[0]]),
            lambda: jnp.array([x[0], x[1]]),
        ],
    )


def _dirac_spinor(p: chex.Array, mass: float, s: int, anti: int):
    """
    Compute the dirac wavefunction.

    Parameters
    ----------
    p:
        Four-momentum of the wavefunction.
    mass:
        Mass of the wavefunction.
    s: int
        Spin of the wavefunction. Must be 1 or -1.
    anti: Int
        If anti = -1, the wavefunction represents a v-spinor.
        If 1, wavefunction represents a u-spinor.
    """
    pm = jnp.linalg.norm(p[1:], axis=0)

    wp = jnp.sqrt(p[0] + pm)
    wm = mass / wp

    w = jax.lax.cond(
        s == -anti,
        lambda: jnp.array([anti * wp, wm]),
        lambda: jnp.array([wm, anti * wp]),
    )

    x = _chi(p, s * anti)

    return jnp.array([w[0] * x[0], w[0] * x[1], w[1] * x[0], w[1] * x[1]])


@functools.partial(jax.vmap, in_axes=(1, None, None), out_axes=1)
def _spinor_u(p: chex.Array, mass: float, s: int):
    return _dirac_spinor(p, mass, s, 1)


@functools.partial(jax.vmap, in_axes=(1, None, None), out_axes=1)
def _spinor_v(p: chex.Array, mass: float, s: int):
    return _dirac_spinor(p, mass, s, -1)


@functools.partial(jax.vmap, in_axes=(1, None, None), out_axes=1)
def _spinor_ubar(p: chex.Array, mass: float, s: int):
    x = jnp.conj(_dirac_spinor(p, mass, s, 1))
    return jnp.array([x[2], x[3], x[0], x[1]])


@functools.partial(jax.vmap, in_axes=(1, None, None), out_axes=1)
def _spinor_vbar(p: chex.Array, mass: float, s: int):
    x = jnp.conj(_dirac_spinor(p, mass, s, -1))
    return jnp.array([x[2], x[3], x[0], x[1]])


def spinor_u(
    momentum: Union[chex.ArrayNumpy, chex.Array], mass: float, spin: int
) -> DiracWf:
    """
    Compute a u-spinor wavefunction.

    Parameters
    ----------
    momentum: ndarray
        Array containing the four-momentum of the particle.
        Must be 1 or 2 dimensional with leading dimension of size 4.
        If 2-dimensional, 2nd dimension must be the batch dimension.
    mass: float
        Mass of the particle.
    spin: int
        Spin of the particle. Must be 1 or -1.
    """
    _check_spin_dirac(spin)
    p: chex.Array = jnp.array(momentum)
    wf = _spinor_u(p, mass, spin)
    return DiracWf(
        wavefunction=wf,
        momentum=p,
        direction=1,
    )


def spinor_v(
    momentum: Union[chex.ArrayNumpy, chex.Array], mass: float, spin: int
) -> DiracWf:
    """
    Compute a v-spinor wavefunction.

    Parameters
    ----------
    momentum: ndarray
        Array containing the four-momentum of the particle.
        Must be 1 or 2 dimensional with leading dimension of size 4.
    mass: float
        Mass of the particle.
    spin: int
        Spin of the particle. Must be 1 or -1.
    """
    _check_spin_dirac(spin)
    p: chex.Array = jnp.array(momentum)
    wf = _spinor_v(p, mass, spin)
    return DiracWf(
        wavefunction=wf,
        momentum=p,
        direction=1,
    )


def spinor_ubar(
    momentum: Union[chex.ArrayNumpy, chex.Array], mass: float, spin: int
) -> DiracWf:
    """
    Compute a ubar-spinor wavefunction.

    Parameters
    ----------
    momentum: ndarray
        Array containing the four-momentum of the particle.
        Must be 1 or 2 dimensional with leading dimension of size 4.
    mass: float
        Mass of the particle.
    spin: int
        Spin of the particle. Must be 1 or -1.
    """
    _check_spin_dirac(spin)
    p: chex.Array = jnp.array(momentum)
    wf = _spinor_ubar(p, mass, spin)
    return DiracWf(
        wavefunction=wf,
        momentum=-p,
        direction=-1,
    )


def spinor_vbar(
    momentum: Union[chex.ArrayNumpy, chex.Array], mass: float, spin: int
) -> DiracWf:
    """
    Compute a vbar-spinor wavefunction.

    Parameters
    ----------
    momentum: ndarray
        Array containing the four-momentum of the particle.
        Must be 1 or 2 dimensional with leading dimension of size 4.
    mass: float
        Mass of the particle.
    spin: int
        Spin of the particle. Must be 1 or -1.
    """
    _check_spin_dirac(spin)
    p: chex.Array = jnp.array(momentum)
    wf = _spinor_vbar(p, mass, spin)
    return DiracWf(
        wavefunction=wf,
        momentum=-p,
        direction=-1,
    )


def charge_conjugate(psi: DiracWf) -> DiracWf:
    """
    Charge conjugate the input wavefunction.

    Parameters
    ----------
    psi: DiracWf
        The dirac wavefunction.

    Returns
    -------
    psi_cc: DiracWf
        Charge conjugated wavefunction.
    """
    s = psi.direction
    wf = jnp.array(
        [
            s * psi.wavefunction[1],
            -s * psi.wavefunction[0],
            -s * psi.wavefunction[3],
            s * psi.wavefunction[2],
        ]
    )
    p = -psi.momentum
    return DiracWf(wavefunction=wf, momentum=p, direction=-s)


# =============================================================================
# ---- Vector Wavefunction ----------------------------------------------------
# =============================================================================


def _check_spin_vector(s: int) -> None:
    assert s in [-1, 0, 1], "Spin must be -1, 0, or 1."


def _polvec_transverse(k: chex.Array, spin: int, s: int) -> chex.Array:
    """
    Compute a transverse (spin 1 or -1) vector wavefunction.

    Parameters
    ----------
    k: ndarray
        Array containing the four-momentum of the particle.
        Must be 1 or 2 dimensional with leading dimension of size 4.
    spin: int
        Spin of the particle. Must be -1, 0, or -1.
    s: int
        If 1, the returned wavefunction is outgoing. If -1, the
        returned wavefunction is incoming. Must be 1 or -1.
    """
    # assert s == 1 or s == -1, "`s` value must be 1 or -1."
    kx, ky, kz = k[1:]
    kt = jnp.hypot(kx, ky)

    def aligned():
        return jnp.array(
            [
                0.0j,
                -spin / jnp.sqrt(2) + 0.0j,
                -jnp.copysign(1.0, kz) * 1.0j / jnp.sqrt(2),
                0.0j,
            ]
        )

    def unaligned():
        km = jnp.sqrt(jnp.square(kx) + jnp.square(ky) + jnp.square(kz))

        kxt = kx / kt / jnp.sqrt(2)
        kyt = ky / kt / jnp.sqrt(2)
        kzm = kz / km
        ktm = kt / km / jnp.sqrt(2)

        eps_0 = 0.0 + 0.0 * 1j
        eps_x = -spin * kxt * kzm + +s * kyt * 1j
        eps_y = -spin * kyt * kzm + -s * kxt * 1j
        eps_z = +spin * ktm + 0.0 * 1j

        return jnp.array([eps_0, eps_x, eps_y, eps_z])

    return jax.lax.cond(kt == 0.0, aligned, unaligned)


def _polvec_longitudinal(k: chex.Array, mass: float) -> chex.Array:
    """
    Compute a longitudinal (spin 0) vector wavefunction.

    Parameters
    ----------
    k: ndarray
        Array containing the four-momentum of the particle.
        Must be 1 or 2 dimensional with leading dimension of size 4.
    mass: float
        Mass of the particle.
    """
    e, kx, ky, kz = k
    km = jnp.linalg.norm(k[1:], axis=0)
    n = e / (mass * km)

    def rest():
        return jnp.array([0.0j, 0.0j, 0.0j, 1.0 + 0.0j])

    def boosted():
        return jnp.array(
            [km / mass + 0.0j, n * kx + 0.0j, n * ky + 0.0j, n * kz + 0.0j]
        )

    def massive():
        return jax.lax.cond(km == 0.0, rest, boosted)

    def massless():
        return jnp.array([0.0j, 0.0j, 0.0j, 0.0j])

    return jax.lax.cond(mass == 0.0, massless, massive)


def _vector_wf(k: chex.Array, mass: float, spin: int, s: int):
    """
    Compute a vector wavefunction.

    Parameters
    ----------
    momentum: ndarray
        Array containing the four-momentum of the particle.
        Must be 1 or 2 dimensional with leading dimension of size 4.
    mass: float
        Mass of the particle.
    spin: int
        Spin of the particle. Must be -1, 0, or -1.
    s: int
        If 1, the returned wavefunction is outgoing. If -1, the
        returned wavefunction is incoming. Must be 1 or -1.
    """
    # assert s == 1 or s == -1, "`s` value must be 1 (incoming) or -1 (outgoing)."
    return jax.lax.switch(
        spin + 1,
        [
            lambda: _polvec_transverse(k, spin, s),
            lambda: _polvec_longitudinal(k, mass),
            lambda: _polvec_transverse(k, spin, s),
        ],
    )


@functools.partial(jax.vmap, in_axes=(1, None, None, None), out_axes=1)
def _vector_wf_vec(k: chex.Array, mass: float, spin: int, s: int):
    return _vector_wf(k, mass, spin, s)


def vector_wf(
    momentum: Union[chex.Array, chex.ArrayNumpy], mass: float, spin: int, out: bool
) -> VectorWf:
    """
    Compute a vector wavefunction.

    Parameters
    ----------
    momentum: ndarray
        Array containing the four-momentum of the particle.
        Must be 1 or 2 dimensional with leading dimension of size 4.
    mass: float
        Mass of the particle.
    spin: int
        Spin of the particle. Must be -1, 0, or -1.
    out: bool
        If true, the returned wavefunction is outgoing.
    """
    _check_spin_vector(spin)
    s = jax.lax.cond(out, lambda: -1, lambda: 1)
    p = jnp.array(momentum)
    wf = _vector_wf_vec(p, mass, spin, s)
    return VectorWf(wavefunction=wf, momentum=s * p, direction=s)


# =============================================================================
# ---- Scalar Wavefunction ----------------------------------------------------
# =============================================================================


def _scalar_wf(k: chex.Array):
    """
    Compute a vector wavefunction.

    Parameters
    ----------
    momentum: ndarray
        Array containing the four-momentum of the particle.
        Must be 1 or 2 dimensional with leading dimension of size 4.
    """
    return jnp.array([1.0], dtype=k.dtype) + 0.0j


@functools.partial(jax.vmap, in_axes=1, out_axes=1)
def _scalar_wf_vec(k: chex.Array):
    return _scalar_wf(k)


def scalar_wf(momentum: Union[chex.Array, chex.ArrayNumpy], out: bool) -> ScalarWf:
    """
    Compute a vector wavefunction.

    Parameters
    ----------
    momentum: ndarray
        Array containing the four-momentum of the particle.
        Must be 1 or 2 dimensional with leading dimension of size 4.
    """
    s = jax.lax.cond(out, lambda: -1, lambda: 1)
    p = jnp.array(momentum)
    wf = _scalar_wf_vec(p)
    return ScalarWf(wavefunction=wf, momentum=s * p, direction=s)



"""
helax (`helax`)

Helax is a package for computing helicity amplitudes. We offer two different
flavors: a pure `NumPy` version and a `jax` version. The `jax` version requires
`jax` and `jaxlib` to be installed. The `jax` version is useful for utilizing a
GPU.
"""



"""
Module for working with Lorentz-invariant phase-space.

"""


import math
from typing import Callable, Dict, Iterable, List, Optional, Sequence, Tuple, Union

import numpy as np
import numpy.typing as npt
from scipy import integrate

from .lvector import lnorm_sqr
from .typing import RealArray
from .utils import kallen_lambda

MassList = Union[List[float], RealArray]
SquaredMatrixElement = Callable[[RealArray], Union[float, RealArray]]


def __flat_squared_matrix_element(momentum: RealArray):
    return np.ones((momentum.shape[-1],), dtype=momentum.dtype)


class PhaseSpace:
    """
    Phase space generator and integrator.
    """

    def __init__(
        self,
        cme: float,
        masses: Union[np.ndarray, Sequence[float]],
        msqrd: Optional[SquaredMatrixElement] = None,
    ) -> None:
        """
        Parameters
        ----------
        cme: float
            Center-of-mass energy of the proccess.
        masses: array-like
            Array of the final state particle masses.
        msqrd: Callable[[ndarray], float], optional
            Squared matrix element of the proccess.
        """
        ms = np.array(masses)
        n = len(ms)

        self.__cme = cme
        self.__masses = ms.reshape((n, 1))
        self.__msqrd: Optional[SquaredMatrixElement] = msqrd

        self.__n = n
        self.__xi0 = np.sqrt(1.0 - (np.sum(ms) / cme) ** 2)
        self.__base_wgt = self._compute_base_weight()

        self.__rng = np.random.default_rng()

        self.__eps = 2 ** (-52)

    @property
    def cme(self) -> float:
        """
        Center-of-mass energy of the proccess.
        """
        return self.__cme

    @cme.setter
    def cme(self, val) -> None:
        self.__cme = val
        self.__base_wgt = self._compute_base_weight()

    @property
    def masses(self) -> np.ndarray:
        """
        Masses of the final state particles.
        """
        return self.__masses.squeeze()

    @masses.setter
    def masses(self, masses: npt.NDArray) -> None:
        self.__n = len(masses)
        self.__masses = np.array(masses).reshape((self.__n, 1))
        self.__base_wgt = self._compute_base_weight()

    @property
    def msqrd(self) -> Optional[SquaredMatrixElement]:
        """
        Squared matrix element of the proccess.
        """
        return self.__msqrd

    def _compute_base_weight(self) -> float:
        """
        Compute the starting weight for each phase space point.
        """
        cme = self.__cme
        n = self.__n

        fact_nm2 = math.factorial(n - 2)
        fact = 1.0 / ((n - 1) * fact_nm2**2)

        return (
            fact
            * (0.5 * np.pi) ** (n - 1)
            * cme ** (2 * n - 4)
            * (0.5 / np.pi) ** (3 * n - 4)
        )

    def _initialize_momenta(self, batch_size: int, dtype=np.float64) -> np.ndarray:
        """
        Initialize isotropic and massless momenta with energyies
        distributed according to E exp(-E).

        Parameters
        ----------
        batch_size: int
            Number of phase space points to generate.

        Returns
        -------
        momenta: ndarray
            Array containing the initialized four momenta with shape
            (4, # final state particles, batch_size). The 1st dimension
            contains the energy, x-, y- and z-components of the 3-momenta. The
            2nd dimension runs over all final state particles. The 3rd
            dimension runs over all events.
        """
        n = self.__n

        rho1 = self.__rng.random(size=(n, batch_size), dtype=dtype)
        rho2 = self.__rng.random(size=(n, batch_size), dtype=dtype)
        rho3 = self.__rng.random(size=(n, batch_size), dtype=dtype)
        rho4 = self.__rng.random(size=(n, batch_size), dtype=dtype)

        ctheta = 2 * rho1 - 1.0
        stheta = np.sqrt(1.0 - ctheta**2)
        phi = 2.0 * np.pi * rho2
        e = -np.log(rho3 * rho4)

        return np.array(
            [e, e * stheta * np.cos(phi), e * stheta * np.sin(phi), e * ctheta]
        )

    def _boost(self, ps: np.ndarray) -> np.ndarray:
        """
        Boost momenta into the center-of-mass frame. Input should be the
        result of `_initialize_momenta`.

        Parameters
        ----------
        ps: ndarray
            Momenta generated from `_initialize_momenta`.

        Returns
        -------
        momenta: ndarray
            Momenta in the center-of-mass frame. The shape is idential to
            input shape.
        """
        sum_ps = np.sum(ps, axis=1)
        inv_mass = np.sqrt(
            sum_ps[0] ** 2 - sum_ps[1] ** 2 - sum_ps[2] ** 2 - sum_ps[3] ** 2
        )
        inv_mass = 1.0 / inv_mass

        bx = -inv_mass * sum_ps[1]
        by = -inv_mass * sum_ps[2]
        bz = -inv_mass * sum_ps[3]

        x = self.__cme * inv_mass
        g = sum_ps[0] * inv_mass
        a = 1.0 / (1.0 + g)

        bdotp = bx * ps[1] + by * ps[2] + bz * ps[3]
        fact = a * bdotp + ps[0]

        ps[0] = x * (g * ps[0] + bdotp)
        ps[1] = x * (fact * bx + ps[1])
        ps[2] = x * (fact * by + ps[2])
        ps[3] = x * (fact * bz + ps[3])

        return ps

    def _compute_scale_factor(self, ps: np.ndarray, iterations=5) -> np.ndarray:
        """
        Compute the scale factor needed to correct the masses of the final
        state particle momenta.

        Parameters
        ----------
        ps: ndarray
            Momenta returned from `_boost`.

        Returns
        -------
        xi: ndarray
            Scale factors for each phase-space point. The shape is
            (batch_size,) where `batch_size` is the size of the last
            dimension of `ps`.
        """
        cme = self.__cme
        xi0 = self.__xi0

        shape = ps.shape[1:]
        e = ps[0]
        xi = xi0 * np.ones((shape[-1],))
        it = 0
        last = False

        while not last:
            deltaf = np.hypot(e * xi, self.__masses)
            f = np.sum(deltaf, axis=0) - cme
            df = np.sum(xi * e**2 / deltaf, axis=0)
            xi = xi - f / df

            # Newton iterations have quadratic convergence. If we are below sqrt(eps),
            # do only one more iteration
            if np.max(f) < np.sqrt(self.__eps) or it == iterations - 1:
                last = True

            it = it + 1

        return xi

    def _correct_masses(self, ps: np.ndarray) -> np.ndarray:
        """
        Correct the masses of the final state particles.

        Parameters
        ----------
        ps: ndarray
            Momenta returned from `_boost`.

        Returns
        -------
        ps: ndarray
            Momenta with corrected masses. The shape is identical to input.
        """
        xi = self._compute_scale_factor(ps)
        ps[0] = np.hypot(xi * ps[0], self.__masses)
        ps[1:] = xi * ps[1:]
        return ps

    def _compute_weights(self, ps) -> np.ndarray:
        """
        Compute the weights of phase space points.

        Parameters
        ----------
        ps: ndarray
            Momenta returned from `_correct_masses`.

        Returns
        -------
        weights: ndarray
            Weights of the phase space points. The shape is (batch_size, )
            where `batch_size` is the size of the last dimension of `ps`.
        """
        cme = self.__cme
        n = self.__n

        modsqr = np.sum(ps[1:] ** 2, axis=0)
        mod = np.sqrt(modsqr)

        t1 = np.sum(mod / cme, axis=0) ** (2 * n - 3)
        t2 = np.sum(modsqr / ps[0], axis=0)
        t3 = np.prod(mod / ps[0], axis=0)

        return t1 / t2 * t3 * self.__cme * self.__base_wgt

    def generator(
        self, n, batch_size: int, seed: Optional[int] = None, dtype=np.float64
    ) -> Iterable[Tuple[np.ndarray, np.ndarray]]:
        """
        Create a generator the yields four-momenta and weights distributed
        according to Lorentz-invariant phase space.

        Parameters
        ----------
        n: int
            Number of phase space points to generate.
        batch_size: int
            Number of phase space points to generate each step.
        seed: int, optional
            Seed used for numpy random number generator.
        dtype: DTypeLike, optional
            Type used for generation of momenta and weights.

        Yields
        -------
        momenta: ndarray
            Batch of momenta containing the initialized four momenta with shape
            (4, # final state particles, batch_size). The 1st dimension
            contains the energy, x-, y- and z-components of the 3-momenta. The
            2nd dimension runs over all final state particles. The 3rd
            dimension runs over all events.
        weights: ndarray
            Batch of weights of the phase space points. Shape is (batch_size,).

        Examples
        --------
        Integrating over phase-space in batches:

        >>> phase_space = PhaseSpace(cme=10.0, masses=[1.0, 2.0, 3.0])
        >>> integrals = []
        >>> for _, weights in phase_space.generator(n=100, batch_size=10, seed=1234):
        ...     integrals.append(np.nanmean(integrals))
        >>> np.average(integrals)
        0.003632349411041629
        """
        if dtype == np.float64:
            self.__eps = 2 ** (-52)
        elif dtype == np.float32:
            self.__eps = 2 ** (-23)

        self.__rng = np.random.default_rng(seed)

        niters = n // batch_size
        niters += 0 if batch_size * niters == n else 1

        for _ in range(niters):
            ps = self._initialize_momenta(batch_size, dtype=dtype)
            ps = self._boost(ps)
            ps = self._correct_masses(ps)
            ws = self._compute_weights(ps)

            if self.__msqrd is not None:
                ws *= self.__msqrd(ps)
            yield (ps, ws)

    def generate(
        self, n, seed: Optional[int] = None, dtype=np.float64
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Generate four-momenta and weights distributed according to
        Lorentz-invariant phase space.

        Parameters
        ----------
        n: int
            Number of phase space points to generate.
        seed: int, optional
            Seed used for numpy random number generator.
        dtype: DTypeLike, optional
            Type used for generation of momenta and weights.

        Returns
        -------
        momenta: ndarray
            Array containing the initialized four momenta with shape
            (4, # final state particles, n). The 1st dimension
            contains the energy, x-, y- and z-components of the 3-momenta. The
            2nd dimension runs over all final state particles. The 3rd
            dimension runs over all events.
        weights: ndarray
            Array containing weights of the phase space points. Shape is given by
            (n,).

        Examples
        --------
        Integrating over phase-space:

        >>> phase_space = PhaseSpace(cme=10.0, masses=[1.0, 2.0, 3.0])
        >>> integrals = []
        >>> momenta, weights = phase_space.generator(n=100, seed=1234)
        >>> np.average(weights)
        0.0037198111038192366
        """
        if dtype == np.float64:
            self.__eps = 2 ** (-52)
        elif dtype == np.float32:
            self.__eps = 2 ** (-23)

        self.__rng = np.random.default_rng(seed)

        ps = self._initialize_momenta(n, dtype=dtype)
        ps = self._boost(ps)
        ps = self._correct_masses(ps)
        ws = self._compute_weights(ps)

        if self.__msqrd is not None:
            ws *= self.__msqrd(ps)

        return ps, ws

    def _integrate_two_body(self) -> Tuple[float, float]:
        cme = self.cme
        m1, m2 = self.masses
        p = np.sqrt(kallen_lambda(cme**2, m1**2, m2**2)) / (2 * cme)
        e1 = np.hypot(m1, p)
        e2 = np.hypot(m2, p)
        ps = np.zeros((4, 2), dtype=np.float64)

        msqrd = __flat_squared_matrix_element if self.msqrd is None else self.msqrd

        def integrand(z):
            sin = np.sqrt(1 - z**2)
            ps[:, 0] = np.array([e1, sin * p, 0.0, z * p])
            ps[:, 1] = np.array([e2, -sin * p, 0.0, -z * p])
            return msqrd(ps)

        pre = 1.0 / (8.0 * np.pi) * p / cme

        integral, error = integrate.quad(integrand, -1.0, 1.0)

        return integral * pre, error * pre

    def integrate(
        self,
        n: int,
        batch_size: Optional[int] = None,
        seed: Optional[int] = None,
        dtype=np.float64,
    ) -> Tuple[np.float64, np.float64]:
        """
        Integrate over phase space.

        Parameters
        ----------
        n: int
            Number of phase space points used in integration.
        batch_size: int, optional
            If not None, the phase-space integration will be broken up into
            batches, processing `batch_size` points at a time. Default is None.
        seed: int, optional
            Seed used for numpy random number generator.
        dtype: DTypeLike, optional
            Type used for generation of momenta and weights.

        Returns
        -------
        integral: float
            Value of the phase space integration.
        error_estimate: float
            Estimation of the error.
        """
        if batch_size is not None and not batch_size == n:
            integrals = []
            errors = []
            for _, ws in self.generator(n, batch_size, seed, dtype):
                avg = np.nanmean(ws, dtype=np.float64)
                std = np.nanstd(ws, dtype=np.float64, ddof=1) / np.sqrt(batch_size)
                integrals.append(avg)
                errors.append(std)

            # Average of averages okay since all samples are same size
            integral = np.nanmean(integrals)
            # Combined error estimate using quadrature
            error = np.sqrt(np.nansum(np.square(errors))) / len(errors)

            return integral, error

        _, weights = self.generate(n, seed=seed, dtype=dtype)
        integral = np.nanmean(weights, dtype=np.float64)
        error = np.nanstd(weights, dtype=np.float64, ddof=1) / np.sqrt(n)

        return integral, error

    def decay_width(
        self,
        n,
        batch_size: Optional[int] = None,
        seed: Optional[int] = None,
        dtype=np.float64,
    ) -> Tuple[np.float64, np.float64]:
        """
        Compute the decay width.

        Parameters
        ----------
        n: int
            Number of phase space points used in integration.
        seed: int, optional
            Seed used for numpy random number generator.
        dtype: DTypeLike, optional
            Type used for generation of momenta and weights.

        Returns
        -------
        width: float
            Estimated value of the decay width.
        error_estimate: float
            Estimation of the error.
        """
        integral, std = self.integrate(n, batch_size=batch_size, seed=seed, dtype=dtype)

        pre = 0.5 / self.__cme
        width = pre * integral
        error = pre * std

        return width, error

    def cross_section(
        self,
        m1,
        m2,
        n,
        batch_size: Optional[int] = None,
        seed: Optional[int] = None,
        dtype=np.float64,
    ) -> Tuple[float, float]:
        """
        Compute the cross section given input masses.

        Parameters
        ----------
        m1: float
            Mass of the 1st incoming particle.
        m2: float
            Mass of the 2nd incoming particle.
        n: int
            Number of phase space points used in integration.
        seed: int, optional
            Seed used for numpy random number generator.
        dtype: DTypeLike, optional
            Type used for generation of momenta and weights.

        Returns
        -------
        cross_section: float
            Estimated value of the cross-section.
        error_estimate: float
            Estimation of the error.
        """
        integral, std = self.integrate(n, batch_size=batch_size, seed=seed, dtype=dtype)

        pre = 0.5 / np.sqrt(kallen_lambda(self.cme**2, m1**2, m2**2))
        cross_section = pre * integral
        error = pre * std

        return cross_section, error

    def energy_distributions(
        self,
        n: int,
        nbins: int,
        batch_size: Optional[int] = None,
        seed: Optional[int] = None,
        dtype=np.float64,
        keep_edges: bool = False,
    ) -> List[Tuple[RealArray, RealArray]]:
        """
        Generate energy distributions of the final state particles.

        Parameters
        ----------
        n: int
            Number of phase space points used in generating the distributions.
        nbins: int
            Number of bins to use for the distributions.
        batch_size: int, optional
            If not None, the phase-space integration will be broken up into
            batches, processing `batch_size` points at a time. Default is None.
        seed: int, optional
            Seed used for numpy random number generator.
        dtype: DTypeLike, optional
            Type used for generation of momenta and weights.

        Returns
        -------
        distributions: List[Tuple[np.ndarray, np.ndarray]]
            The energy distributions. Each entry in the returned list is a
            tuple with the first item the probability distribution
            (shape=(nbins,)) and the second the energy bin edges
            (shape=(nbins+1,)).
        """
        cme = self.cme
        mass_sum = np.sum(self.masses)

        def bounds(m):
            msum = mass_sum - m
            emin = m
            emax = (cme**2 + m**2 - msum**2) / (2 * cme)
            return emin, emax

        def normalize(probabilities, edges):
            norm = np.sum(
                [p * (edges[i + 1] - edges[i]) for i, p in enumerate(probabilities)]
            )
            return probabilities / norm

        ebounds = [bounds(m) for m in self.masses]
        bins = [np.linspace(emin, emax, nbins + 1) for emin, emax in ebounds]

        distributions = [
            np.histogram(np.zeros(nbins, dtype=np.float64), bins=b)[0] for b in bins
        ]

        if batch_size is not None and not batch_size == n:
            # Build up the distributions from batches
            for ps, ws in self.generator(n, batch_size, seed, dtype):
                for i in range(len(self.masses)):
                    d = np.histogram(ps[0, i], bins=bins[i], weights=ws)[0]
                    distributions[i] += d
        else:
            ps, ws = self.generate(n, seed=seed, dtype=dtype)
            for i in range(len(self.masses)):
                distributions[i] = np.histogram(ps[0, i], bins=bins[i], weights=ws)[0]

        for i in range(len(self.masses)):
            distributions[i] = normalize(distributions[i], bins[i])

        if keep_edges:
            return [(dpde, e) for dpde, e in zip(distributions, bins)]

        centers = [0.5 * (b[1:] + b[:-1]) for b in bins]
        return [(dpde, c) for dpde, c in zip(distributions, centers)]

    def invariant_mass_distributions(
        self,
        n: int,
        nbins: int,
        pairs: Optional[List[Tuple[int, int]]] = None,
        batch_size: Optional[int] = None,
        seed: Optional[int] = None,
        dtype=np.float64,
        keep_edges: bool = False,
    ) -> Dict[Tuple[int, int], Tuple[RealArray, RealArray]]:
        """
        Generate invariant mass distributions of the final state particles.

        The invariant masses are defined as sqrt((pi + pj)^2), where pi and pj
        are the four-momenta of the ith and jth particles.

        Parameters
        ----------
        n: int
            Number of phase space points used in generating the distributions.
        nbins: int
            Number of bins to use for the distributions.
        batch_size: int, optional
            If not None, the phase-space integration will be broken up into
            batches, processing `batch_size` points at a time. Default is None.
        seed: int, optional
            Seed used for numpy random number generator.
        dtype: DTypeLike, optional
            Type used for generation of momenta and weights.

        Returns
        -------
        distributions: Dict[Tuple[int,int], Tuple[np.ndarray, np.ndarray]]
            The invariant mass distributions. Each entry in the returned
            dictionary is a tuple with the first item the probability
            distribution (shape=(nbins,)) and the second the bin edges
            (shape=(nbins+1,)). The keys are the pairs of indices specifying
            which two particles the invariant mass-squared distribution
            corresponds to.
        """
        cme = self.cme
        masses = self.masses
        mass_sum = np.sum(masses)
        nfsp = len(masses)

        if pairs is None:
            pairs_: List[Tuple[int, int]] = []
            for i in range(nfsp):
                for j in range(i + 1, nfsp):
                    pairs_.append((i, j))
        else:
            pairs_ = pairs

        def bounds(m1, m2):
            msum = mass_sum - m1 - m2
            mmin = m1 + m2
            mmax = cme - msum
            return mmin, mmax

        def normalize(probabilities, edges):
            norm = np.sum(
                [p * (edges[i + 1] - edges[i]) for i, p in enumerate(probabilities)]
            )
            return probabilities / norm

        def make_bins(pair):
            mmin, mmax = bounds(masses[pair[0]], masses[pair[1]])
            return np.linspace(mmin, mmax, nbins + 1)

        def inv_mass(ps, pair):
            i, j = pair
            return np.sqrt(np.abs(lnorm_sqr(ps[:, i] + ps[:, j])))

        bins = [make_bins(pair) for pair in pairs_]

        distributions = [
            np.histogram(np.zeros(nbins, dtype=np.float64), bins=b)[0] for b in bins
        ]

        if batch_size is not None and not batch_size == n:
            # Build up the distributions from batches
            for ps, ws in self.generator(n, batch_size, seed, dtype):
                for i, pair in enumerate(pairs_):
                    d = np.histogram(inv_mass(ps, pair), bins=bins[i], weights=ws)[0]
                    distributions[i] += d
        else:
            ps, ws = self.generate(n, seed=seed, dtype=dtype)
            for i, pair in enumerate(pairs_):
                distributions[i] = np.histogram(
                    inv_mass(ps, pair), bins=bins[i], weights=ws
                )[0]

        for i in range(len(pairs_)):
            distributions[i] = normalize(distributions[i], bins[i])

        if keep_edges:
            return {pair: (distributions[i], bins[i]) for i, pair in enumerate(pairs_)}

        centers = [0.5 * (b[1:] + b[:-1]) for b in bins]
        return {pair: (distributions[i], centers[i]) for i, pair in enumerate(pairs_)}



from typing import Tuple

import numpy as np
import pytest

from helax.numpy import amplitudes, wavefunctions
from helax.numpy.lvector import lnorm_sqr
from helax.numpy.phase_space import PhaseSpace
from helax.numpy.utils import abs2, kallen_lambda
from helax.numpy.wavefunctions import DiracWf, VectorWf
from helax.vertices import VertexFFV

MUON_MASS = 1.056584e-01
G_FERMI = 1.166379e-05
EL = 3.028620e-01
SW = 4.808530e-01
CW = np.sqrt(1.0 - SW**2)
MASS_W = 8.038500e01
WIDTH_W = 2.085000e00
MASS_B = 4.18
MASS_T = 172.9
MASS_MU = 105.6583745e-3
MASS_E = 0.5109989461e-3
MASS_Z = 91.1876

MASS_H = 125.00
WIDTH_H = 0.00374
VEV_H = 246.21965

SEED = 1234


def spinor_u(momenta, mass: float) -> Tuple[DiracWf, DiracWf]:
    return (
        wavefunctions.spinor_u(momenta, mass, -1),
        wavefunctions.spinor_u(momenta, mass, 1),
    )


def spinor_v(momenta, mass: float) -> Tuple[DiracWf, DiracWf]:
    return (
        wavefunctions.spinor_v(momenta, mass, -1),
        wavefunctions.spinor_v(momenta, mass, 1),
    )


def spinor_ubar(momenta, mass: float) -> Tuple[DiracWf, DiracWf]:
    return (
        wavefunctions.spinor_ubar(momenta, mass, -1),
        wavefunctions.spinor_ubar(momenta, mass, 1),
    )


def spinor_vbar(momenta, mass: float) -> Tuple[DiracWf, DiracWf]:
    return (
        wavefunctions.spinor_vbar(momenta, mass, -1),
        wavefunctions.spinor_vbar(momenta, mass, 1),
    )


def polvec(momenta, mass: float, out: bool) -> Tuple[VectorWf, VectorWf, VectorWf]:
    return (
        wavefunctions.vector_wf(momenta, mass, -1, out),
        wavefunctions.vector_wf(momenta, mass, 0, out),
        wavefunctions.vector_wf(momenta, mass, 1, out),
    )


def current_ff_to_v(
    vertex, mass: float, width: float, psi_out: DiracWf, psi_in: DiracWf
):
    return amplitudes.current_ff_to_v(vertex, mass, width, psi_out, psi_in)


def amplitude_ffv(vertex, psi_out: DiracWf, psi_in: DiracWf, eps: VectorWf):
    return amplitudes.amplitude_ffv(vertex, psi_out, psi_in, eps)


def wavefunction_iter(*wavefuncs):
    idx_list = []
    for wf in wavefuncs:
        if isinstance(wf, wavefunctions.ScalarWf):
            idx_list.append([0])
        elif isinstance(wf, wavefunctions.DiracWf):
            idx_list.append([0, 1])
        elif isinstance(wf, wavefunctions.VectorWf):
            idx_list.append([0, 1, 2])
        else:
            raise ValueError(f"Invalid wf {type(wf)}")

    idxs = np.array(np.meshgrid(*idx_list)).T.reshape(-1, len(idx_list))
    for idx in idxs:
        yield tuple([wf[i] for wf, i in zip(wavefuncs, idx)])


def test_width_mu_to_e_nu_nu():
    def msqrd_mu_to_e_nu_nu_analytic(momenta):
        t = lnorm_sqr(momenta[:, 0] + momenta[:, 2])
        s = lnorm_sqr(momenta[:, 1] + momenta[:, 1])

        return (
            -0.125
            * (
                EL**4
                * (
                    MASS_E**4 * MASS_MU**2 * (MASS_MU**2 - s - t)
                    + 4 * MASS_W**4 * t * (-(MASS_MU**2) + t)
                    + MASS_E**2
                    * (
                        -4 * MASS_W**4 * t
                        - MASS_MU**4 * (s + t)
                        + MASS_MU**2
                        * (4 * MASS_W**4 + 4 * MASS_W**2 * s + (s + t) ** 2)
                    )
                )
            )
            / (
                MASS_W**4
                * SW**4
                * (
                    (-(MASS_E**2) - MASS_MU**2 + MASS_W**2 + s + t) ** 2
                    + MASS_W**2 * WIDTH_W**2
                )
            )
        )

    def msqrd_mu_to_e_nu_nu_hel(momenta):
        pe = momenta[:, 0]
        pve = momenta[:, 1]
        pvm = momenta[:, 2]
        pmu = np.sum(momenta, axis=1)

        v_wll = VertexFFV(left=EL / (np.sqrt(2) * SW), right=0.0)

        mu_wfs = spinor_u(pmu, MUON_MASS)
        e_wfs = spinor_ubar(pe, 0.0)
        ve_wfs = spinor_v(pve, 0.0)
        vm_wfs = spinor_ubar(pvm, 0.0)

        def msqrd(imu, ie, ive, ivm):
            w_wf = current_ff_to_v(v_wll, MASS_W, WIDTH_W, vm_wfs[ivm], mu_wfs[imu])
            amp = amplitude_ffv(v_wll, e_wfs[ie], ve_wfs[ive], w_wf)
            return abs2(amp)

        idxs = np.array(np.meshgrid([0, 1], [0, 1], [0, 1], [0, 1])).T.reshape(-1, 4)
        res = sum(msqrd(i_n, i_v1, i_v2, i_v3) for (i_n, i_v1, i_v2, i_v3) in idxs)

        return res / 2.0

    phase_space = PhaseSpace(MASS_MU, [MASS_E, 0.0, 0.0], msqrd=msqrd_mu_to_e_nu_nu_hel)
    width = phase_space.decay_width(10_000, seed=SEED)[0]

    phase_space = PhaseSpace(
        MASS_MU, [MASS_E, 0.0, 0.0], msqrd=msqrd_mu_to_e_nu_nu_analytic
    )
    width2 = phase_space.decay_width(10_000, seed=SEED)[0]

    assert float(width) == pytest.approx(width2, abs=0.0, rel=0.01)


def test_width_t_to_b_w():
    def msqrd_t_to_b_w(momenta):
        pb = momenta[:, 0]
        pw = momenta[:, 1]
        pt = np.sum(momenta, axis=1)

        t_wfs = spinor_u(pt, MASS_T)
        b_wfs = spinor_ubar(pb, MASS_B)
        w_wfs = polvec(pw, MASS_W, True)

        v_tbw = VertexFFV(left=EL / (np.sqrt(2.0) * SW), right=0.0)

        def msqrd(it, ib, iw):
            amp = amplitude_ffv(v_tbw, b_wfs[ib], t_wfs[it], w_wfs[iw])
            return abs2(amp)

        idxs = np.array(np.meshgrid([0, 1], [0, 1], [0, 1, 2])).T.reshape(-1, 3)
        res = sum(msqrd(it, ib, iw) for (it, ib, iw) in idxs)

        return res / 2.0

    analytic = (
        EL**2
        * (
            (MASS_T**2 + 2 * MASS_W**2) * (MASS_T**2 - MASS_W**2)
            + MASS_B**2 * (MASS_W**2 - 2 * MASS_T**2)
            + MASS_B**4
        )
        * np.sqrt(kallen_lambda(MASS_B**2, MASS_T**2, MASS_W**2))
    ) / (64 * MASS_T**3 * MASS_W**2 * np.pi * SW**2)

    phase_space = PhaseSpace(MASS_T, [MASS_B, MASS_W], msqrd=msqrd_t_to_b_w)
    width = phase_space.decay_width(10_000, seed=SEED)[0]

    assert float(width) == pytest.approx(analytic, rel=1e-3, abs=0.0)


# ============================================================================
# ---- chi + chi -> V -> f + f -----------------------------------------------
# ============================================================================


def test_dark_matter_annihilation_vector_mediator():
    gvxx = 1.0
    gvff = 1.0
    mx = 10.0
    mf = 1.0
    cme = 3 * mx
    mv = 1.0
    widthv = 1.0

    vxx = VertexFFV(left=gvxx, right=gvxx)
    vff = VertexFFV(left=gvff, right=gvff)

    ex = cme / 2.0
    px = np.sqrt(ex**2 - mx**2)
    p1 = np.expand_dims(np.array([ex, 0.0, 0.0, px]), -1)
    p2 = np.expand_dims(np.array([ex, 0.0, 0.0, -px]), -1)

    wf1s = spinor_u(p1, mx)
    wf2s = spinor_vbar(p2, mx)

    def msqrd_x_x_to_f_f(momenta):
        p3 = momenta[:, 0]
        p4 = momenta[:, 1]

        wf3s = spinor_ubar(p3, mf)
        wf4s = spinor_v(p4, mf)

        def msqrd(i1, i2, i3, i4):
            vwf = current_ff_to_v(vxx, mv, widthv, wf2s[i2], wf1s[i1])
            return abs2(amplitude_ffv(vff, wf3s[i3], wf4s[i4], vwf))

        idxs = np.array(np.meshgrid([0, 1], [0, 1], [0, 1], [0, 1])).T.reshape(-1, 4)
        res = sum(msqrd(i1, i2, i3, i4) for (i1, i2, i3, i4) in idxs)

        return res / 4.0

    phase_space = PhaseSpace(cme, [mf, mf], msqrd=msqrd_x_x_to_f_f)
    cs = phase_space.cross_section(mx, mx, 10_000, seed=SEED)[0]

    s = cme**2
    analytic = (
        gvff**2
        * gvxx**2
        * np.sqrt(s * (-4 * mf**2 + s))
        * (2 * mf**2 + s)
        * (2 * mx**2 + s)
    ) / (
        12.0
        * np.pi
        * s
        * np.sqrt(s * (-4 * mx**2 + s))
        * (mv**4 + s**2 + mv**2 * (-2 * s + widthv**2))
    )

    assert float(cs) == pytest.approx(analytic, rel=1e-2, abs=0.0)


# ============================================================================
# ---- Z -> f + f ------------------------------------------------------------
# ============================================================================


def widths_z_to_f_f(mf, ncf, t3f, qf):
    pre = EL / (SW * CW)
    left = pre * (t3f - qf * SW**2)
    right = -pre * qf * SW**2
    vertex = VertexFFV(left=left, right=right)

    def msqrd_z_to_f_f(momenta):
        pf1 = momenta[:, 0]
        pf2 = momenta[:, 1]
        pz = np.sum(momenta, axis=1)

        z_wfs = polvec(pz, MASS_Z, False)
        ubar = spinor_ubar(pf1, mf)
        v = spinor_v(pf2, mf)

        def msqrd(iz, i1, i2):
            amp = amplitude_ffv(vertex, ubar[i1], v[i2], z_wfs[iz])
            return abs2(amp)

        idxs = np.array(np.meshgrid([0, 1, 2], [0, 1], [0, 1])).T.reshape(-1, 3)
        res = sum(msqrd(iz, i1, i2) for (iz, i1, i2) in idxs)

        return ncf * res / 3.0

    mr2 = (mf / MASS_Z) ** 2
    af = t3f - qf * SW**2
    bf = -qf * SW**2

    analytic = (
        ncf
        * EL**2
        * MASS_Z
        / (24 * np.pi * CW**2 * SW**2)
        * np.sqrt(1 - 4 * mr2)
        * ((af**2 + bf**2) * (1 - mr2) + 6 * af * bf * mr2)
    )

    phase_space = PhaseSpace(MASS_Z, [mf, mf], msqrd=msqrd_z_to_f_f)
    width = phase_space.decay_width(10_000, seed=SEED)[0]

    return width, analytic


def test_width_z_to_b_b():
    mf = MASS_B
    ncf = 3
    t3f = -0.5
    qf = -1.0 / 3.0

    width, analytic = widths_z_to_f_f(mf=mf, ncf=ncf, t3f=t3f, qf=qf)
    assert float(width / analytic) == pytest.approx(1, rel=1e-3, abs=0.0)


def test_width_z_to_e_e():
    mf = MASS_E
    ncf = 1
    t3f = -0.5
    qf = -1.0

    width, analytic = widths_z_to_f_f(mf=mf, ncf=ncf, t3f=t3f, qf=qf)
    assert float(width / analytic) == pytest.approx(1.0, rel=1e-3, abs=0.0)



"""Low-level HELAS functions."""

import numpy as np

# double complex fi[5],chi[1]
# double precision p(0:3),sf[1],sfomeg[1],omega[1],fmass,pp,pp3,sqp0p3,sqm(0:1)
# integer nhel,nsf,ip,im,nh

# double precision 0.0, 0.5, 2.0
# parameter( 0.0 = 0.0, 0.5 = 0.5, 2.0 = 2.0 )
def ixxxxx(p, fmass, nhel, nsf, fi):
    fi[4] = (p[0] + 1j * p[3]) * nsf
    fi[5] = (p[1] + 1j * p[2]) * nsf

    sqm = np.zeros((2,), dtype=np.float64)
    sf = np.zeros((2,), dtype=np.float64)
    sfomeg = np.zeros((2,), dtype=np.float64)
    omega = np.zeros((2,), dtype=np.float64)
    chi = np.zeros((2,), dtype=np.complex128)

    nh = nhel * nsf

    if fmass != 0.0:
        pp = min(p[0], np.sqrt(p[1] ** 2 + p[2] ** 2 + p[3] ** 2))

        if pp == 0.0:
            sqm[0] = np.sqrt(abs(fmass))  # possibility of negative fermion masses
            sqm[1] = sign(sqm[0], fmass)  # possibility of negative fermion masses
            ip = (1 + nh) / 2
            im = (1 - nh) / 2
            fi[0] = ip * sqm[ip]
            fi[1] = im * nsf * sqm[ip]
            fi[2] = ip * nsf * sqm[im]
            fi[3] = im * sqm[im]
        else:
            sf[0] = float(1 + nsf + (1 - nsf) * nh) * 0.5
            sf[1] = float(1 + nsf - (1 - nsf) * nh) * 0.5
            omega[0] = np.sqrt(p[0] + pp)
            omega[1] = fmass / omega[0]
            ip = (3 + nh) / 2
            im = (3 - nh) / 2
            sfomeg[0] = sf[0] * omega[ip]
            sfomeg[1] = sf[1] * omega[im]
            pp3 = max(pp + p[3], 0.0)
            chi[0] = np.sqrt(pp3 * 0.5 / pp)
            if pp3 == 0.0:
                chi[1] = -nh
            else:
                chi[1] = (nh * p[1] + 1j * p[2]) / np.sqrt(2.0 * pp * pp3)
            fi[0] = sfomeg[0] * chi[im]
            fi[1] = sfomeg[0] * chi[ip]
            fi[2] = sfomeg[1] * chi[im]
            fi[3] = sfomeg[1] * chi[ip]
    else:
        if p[1] == 0 and p[2] == 0 and p[3] < 0:
            sqp0p3 = 0
        else:
            sqp0p3 = np.sqrt(max(p[0] + p[3], 0.0)) * nsf

        chi[0] = sqp0p3
        if sqp0p3 == 0.0:
            chi[1] = (-nhel) * np.sqrt(2.0 * p[0])
        else:
            chi[1] = (nh * p[1] + 1j * p[2]) / sqp0p3

        if nh == 1:
            fi[0] = 0.0
            fi[1] = 0.0
            fi[2] = chi[0]
            fi[3] = chi[1]
        else:
            fi[0] = chi[1]
            fi[1] = chi[0]
            fi[2] = 0.0
            fi[3] = 0.0

    return


#
# double complex fo[5],chi[1]
# double precision p(0:3),sf[1],sfomeg[1],omega[1],fmass,pp,pp3,sqp0p3,sqm(0:1)
# integer nhel,nsf,nh,ip,im
# double precision 0.0, 0.5, 2.0
# parameter( 0.0 = 0.0, 0.5 = 0.5, 2.0 = 2.0 )
def oxxxxx(p, fmass, nhel, nsf, fo):
    fo[4] = (p[0] + 1j * p[3]) * nsf
    fo[5] = (p[1] + 1j * p[2]) * nsf
    nh = nhel * nsf
    if fmass != 0.0:
        pp = min(p[0], np.sqrt(p[1] ** 2 + p[2] ** 2 + p[3] ** 2))
        if pp == 0.0:
            sqm[0] = np.sqrt(abs(fmass))  # possibility of negative fermion masses
            sqm[1] = sign(sqm[0], fmass)  # possibility of negative fermion masses
            ip = -((1 + nh) / 2)
            im = (1 - nh) / 2
            fo[0] = im * sqm(im)
            fo[1] = ip * nsf * sqm(im)
            fo[2] = im * nsf * sqm(-ip)
            fo[3] = ip * sqm(-ip)
        else:
            pp = min(p[0], np.sqrt(p[1] ** 2 + p[2] ** 2 + p[3] ** 2))
            sf[0] = dble(1 + nsf + (1 - nsf) * nh) * 0.5
            sf[1] = dble(1 + nsf - (1 - nsf) * nh) * 0.5
            omega[0] = np.sqrt(p[0] + pp)
            omega[1] = fmass / omega[0]
            ip = (3 + nh) / 2
            im = (3 - nh) / 2
            sfomeg[0] = sf[0] * omega[ip]
            sfomeg[1] = sf[1] * omega[im]
            pp3 = max(pp + p[3], 0.0)
            chi[0] = np.sqrt(pp3 * 0.5 / pp)
            if pp3 == 0.0:
                chi[1] = -nh
            else:
                chi[1] = (nh * p[1] + 1j * -p[2]) / np.sqrt(2.0 * pp * pp3)
            fo[0] = sfomeg[1] * chi[im]
            fo[1] = sfomeg[1] * chi[ip]
            fo[2] = sfomeg[0] * chi[im]
            fo[3] = sfomeg[0] * chi[ip]
    else:
        if p[1] == 0 and p[2] == 0 and p[3] < 0:
            sqp0p3 = 0
        else:
            sqp0p3 = np.sqrt(max(p[0] + p[3], 0.0)) * nsf
        chi[0] = sqp0p3
        if sqp0p3 == 0.0:
            chi[1] = (-nhel) * np.sqrt(2.0 * p[0])
        else:
            chi[1] = (nh * p[1] + 1j * -p[2]) / sqp0p3
        if nh == 1:
            fo[0] = chi[0]
            fo[1] = chi[1]
            fo[2] = 0.0
            fo[3] = 0.0
        else:
            fo[0] = 0.0
            fo[1] = 0.0
            fo[2] = chi[1]
            fo[3] = chi[0]
    return


#
# This subroutine computes a VECTOR wavefunction.
#
# input:
#       real    p(0:3)         : four-momentum of vector boson
#       real    vmass          : mass          of vector boson
#       integer nhel = -1, 0, 1: helicity      of vector boson
#                                (0 is forbidden if vmass=0.0)
#       integer nsv  = -1 or 1 : +1 for final, -1 for initial
#
# output:
#       complex vc[5]          : vector wavefunction       epsilon^mu(v)
#
# double complex vc[5]
# double precision p(0:3),vmass,hel,hel0,pt,pt2,pp,pzpt,emp,sqh
# integer nhel,nsv,nsvahl

# double precision 0.0, 0.5, 1.0, 2.0
# parameter( 0.0 = 0.0, 0.5 = 0.5 )
# parameter( 1.0 = 1.0, 2.0 = 2.0 )
def vxxxxx(p, vmass, nhel, nsv, vc):
    sqh = np.sqrt(0.5)
    hel = float(nhel)
    nsvahl = nsv * abs(hel)
    pt2 = p[1] ** 2 + p[2] ** 2
    pp = min(p[0], np.sqrt(pt2 + p[3] ** 2))
    pt = min(pp, np.sqrt(pt2))
    vc[4] = (p[0] + 1j * p[3]) * nsv
    vc[5] = (p[1] + 1j * p[2]) * nsv
    if vmass != 0.0:
        hel0 = 1.0 - dabs(hel)
        if pp == 0.0:
            vc[0] = complex(0.0)
            vc[1] = complex(-hel * sqh)
            vc[2] = complex(0.0 + 1j * nsvahl * sqh)
            vc[3] = complex(hel0)
        else:
            emp = p[0] / (vmass * pp)
            vc[0] = complex(hel0 * pp / vmass)
            vc[3] = complex(hel0 * p[3] * emp + hel * pt / pp * sqh)
            if pt != 0.0:
                pzpt = p[3] / (pp * pt) * sqh * hel
                vc[1] = complex(
                    hel0 * p[1] * emp - p[1] * pzpt,
                    -nsvahl * p[2] / pt * sqh,
                )
                vc[2] = complex(
                    hel0 * p[2] * emp - p[2] * pzpt,
                    nsvahl * p[1] / pt * sqh,
                )
            else:
                vc[1] = complex(-hel * sqh)
                vc[2] = complex(0.0, nsvahl * sign(sqh, p[3]))
    else:
        pp = p[0]
        pt = np.sqrt(p[1] ** 2 + p[2] ** 2)
        vc[0] = complex(0.0)
        vc[3] = complex(hel * pt / pp * sqh)
        if pt != 0.0:
            pzpt = p[3] / (pp * pt) * sqh * hel
            vc[1] = complex(-p[1] * pzpt, -nsv * p[2] / pt * sqh)
            vc[2] = complex(-p[2] * pzpt, nsv * p[1] / pt * sqh)
        else:
            vc[1] = complex(-hel * sqh)
            vc[2] = complex(0.0, nsv * sign(sqh, p[3]))
    return


#
# This subroutine computes a complex SCALAR wavefunction.
#
# input:
#       real    p(0:3)         : four-momentum of scalar boson
#       integer nss  = -1 or 1 : +1 for final, -1 for initial
#
# output:
#       complex sc[2]          : scalar wavefunction                   s
#
# double complex sc[2]
# double precision p(0:3)
# integer nss

# double precision 1.0
# parameter( 1.0 = 1.0 )
def sxxxxx(p, nss, sc):
    sc[0] = complex(1.0)
    sc[1] = complex(p[0], p[3]) * nss
    sc[2] = complex(p[1], p[2]) * nss
    return


#
# This subroutine computes an amplitude of the fermion-fermion-vector
# coupling.
#
# input:
#       complex fi[5]          : flow-in  fermion                   |fi>
#       complex fo[5]          : flow-out fermion                   <fo|
#       complex vc[5]          : input    vector                      v
#       complex gc[1]          : coupling constants                  gvf
#
# output:
#       complex vertex         : amplitude                     <fo|v|fi>
#
# double complex fi[5],fo[5],gc[1],vc[5],vertex

# double precision 0.0, 1.0
# parameter( 0.0 = 0.0 )
# double complex cImag, cZero
# parameter( cImag = ( 0.0, 1.0 ), cZero = ( 0.0, 0.0 ) )
def iovxxx(fi, fo, vc, gc, vertex):
    vertex = gc[0] * (
        (fo[2] * fi[0] + fo[3] * fi[1]) * vc[0]
        + (fo[2] * fi[1] + fo[3] * fi[0]) * vc[1]
        - (fo[2] * fi[1] - fo[3] * fi[0]) * vc[2] * cImag
        + (fo[2] * fi[0] - fo[3] * fi[1]) * vc[3]
    )
    if gc[1] != cZero:
        vertex = vertex + gc[1] * (
            (fo[0] * fi[2] + fo[1] * fi[3]) * vc[0]
            - (fo[0] * fi[3] + fo[1] * fi[2]) * vc[1]
            + (fo[0] * fi[3] - fo[1] * fi[2]) * vc[2] * cImag
            - (fo[0] * fi[2] - fo[1] * fi[3]) * vc[3]
        )
    return


#
# This subroutine computes the sum of photon and Z currents with the
# suitable weights ( j(W3) = cos(theta_W) j(Z) + sin(theta_W) j(A) ).
# The output j3 is useful as an input of vvvxxx, jvvxxx or w3w3xx.
# The photon propagator is given in Feynman gauge, and the Z propagator
# is given in unitary gauge.
#
# input:
#       complex fi[5]          : flow-in  fermion                   |fi>
#       complex fo[5]          : flow-out fermion                   <fo|
#       complex gaf[1]         : fi couplings with A                 gaf
#       complex gzf[1]         : fi couplings with Z                 gzf
#       real    zmass          : mass  of Z
#       real    zwidth         : width of Z
#
# output:
#       complex j3[5]          : W3 current             j^mu(<fo|w3|fi>)
#
# double complex fi[5],fo[5],j3[5],gaf[1],gzf[1]
# double complex c0l,c1l,c2l,c3l,csl,c0r,c1r,c2r,c3r,csr,dz,ddif
# double complex gn,gz3l,ga3l
# double complex cm2  ! mass**2- I Gamma mass (Fabio)
# double precision q(0:3),zmass,zwidth,zm2,zmw
# double precision q2,da,ww,cw,sw

# double precision 0.0, 1.0
# parameter( 0.0 = 0.0, 1.0 = 1.0 )
# double complex cImag, cZero
# parameter( cImag = ( 0.0, 1.0 ), cZero = ( 0.0, 0.0 ) )
def j3xxxx(fi, fo, gaf, gzf, zmass, zwidth, j3):
    j3[4] = fo[4] - fi[4]
    j3[5] = fo[5] - fi[5]
    q[0] = -dble(j3[4])
    q[1] = -dble(j3[5])
    q[2] = -dimag(j3[5])
    q[3] = -dimag(j3[4])
    q2 = q[0] ** 2 - (q[1] ** 2 + q[2] ** 2 + q[3] ** 2)
    zm2 = zmass**2
    zmw = zmass * zwidth
    da = 1.0 / q2
    # ww = max(dsign(zmw,q2), 0.0)
    dz = 1.0 / (q2 - zm2 + 1j * zmw)
    ddif = (-zm2 + 1j * zmw) * da * dz

    # ddif is the difference : ddif=da-dz
    #  For the running width, use below instead of the above ww,dz and ddif.
    #      ww = max( zwidth*q2/zmass, 0.0 )
    #      dz = 1.0/dcmplx( q2-zm2, zmw )
    #      ddif = dcmplx( -zm2, zmw )*da*dz
    cw = 1.0 / sqrt(1.0 + (gzf[1] / gaf[1]) ** 2)
    sw = sqrt((1.0 - cw) * (1.0 + cw))
    gn = gaf[1] * sw
    gz3l = gzf[0] * cw
    ga3l = gaf[0] * sw
    c0l = fo[2] * fi[0] + fo[3] * fi[1]
    c0r = fo[0] * fi[2] + fo[1] * fi[3]
    c1l = -(fo[2] * fi[1] + fo[3] * fi[0])
    c1r = fo[0] * fi[3] + fo[1] * fi[2]
    c2l = (fo[2] * fi[1] - fo[3] * fi[0]) * cImag
    c2r = (-fo[0] * fi[3] + fo[1] * fi[2]) * cImag
    c3l = -fo[2] * fi[0] + fo[3] * fi[1]
    c3r = fo[0] * fi[2] - fo[1] * fi[3]
    # Fabio's implementation of the fixed width
    cm2 = zm2 + 1j * -zmw
    #     csl = (q[0+1]*c0l-q[1+1]*c1l-q[2+1]*c2l-q[3+1]*c3l)/zm2
    #     csr = (q[0+1]*c0r-q[1+1]*c1r-q[2+1]*c2r-q[3+1]*c3r)/zm2
    csl = (q[0] * c0l - q[1] * c1l - q[2] * c2l - q[3] * c3l) / cm2
    csr = (q[0] * c0r - q[1] * c1r - q[2] * c2r - q[3] * c3r) / cm2

    j3[0] = (
        gz3l * dz * (c0l - csl * q[0])
        + ga3l * c0l * da
        + gn * (c0r * ddif + csr * q[0] * dz)
    )
    j3[1] = (
        gz3l * dz * (c1l - csl * q[1])
        + ga3l * c1l * da
        + gn * (c1r * ddif + csr * q[1] * dz)
    )
    j3[2] = (
        gz3l * dz * (c2l - csl * q[2])
        + ga3l * c2l * da
        + gn * (c2r * ddif + csr * q[2] * dz)
    )
    j3[3] = (
        gz3l * dz * (c3l - csl * q[3])
        + ga3l * c3l * da
        + gn * (c3r * ddif + csr * q[3] * dz)
    )
    return


#
# This subroutine computes an off-shell fermion wavefunction from a
# flowing-OUT external fermion and a vector boson.
#
#
# input:
#       complex fo[5]          : flow-out fermion                   <fo|
#       complex vc[5]          : input    vector                      v
#       complex gc[1]          : coupling constants                  gvf
#       real    fmass          : mass  of OUTPUT fermion f'
#       real    fwidth         : width of OUTPUT fermion f'
#
# output:
#       complex fvo[5]         : off-shell fermion             <fo,v,f'|
#
# double complex fo[5],vc[5],gc[1],fvo[5],sl1,sl2,sr1,sr2,d
# double precision pf(0:3),fmass,fwidth,pf2

# double precision 0.0, 1.0
# parameter( 0.0 = 0.0, 1.0 = 1.0 )
# double complex cImag, cZero
# parameter( cImag = ( 0.0, 1.0 ), cZero = ( 0.0, 0.0 ) )
def fvoxxx(fo, vc, gc, fmass, fwidth, fvo):
    fvo[4] = fo[4] + vc[4]
    fvo[5] = fo[5] + vc[5]
    pf[0] = dble(fvo[4])
    pf[1] = dble(fvo[5])
    pf[2] = dimag(fvo[5])
    pf[3] = dimag(fvo[4])
    pf2 = pf[0] ** 2 - (pf[1] ** 2 + pf[2] ** 2 + pf[3] ** 2)
    d = -1.0 / (pf2 - fmass**2 + 1j * fmass * fwidth)
    sl1 = (vc[0] + vc[3]) * fo[2] + (vc[1] + cImag * vc[2]) * fo[3]
    sl2 = (vc[1] - cImag * vc[2]) * fo[2] + (vc[0] - vc[3]) * fo[3]
    if gc[1] != cZero:
        sr1 = (vc[0] - vc[3]) * fo[0] - (vc[1] + cImag * vc[2]) * fo[1]
        sr2 = -(vc[1] - cImag * vc[2]) * fo[0] + (vc[0] + vc[3]) * fo[1]
        fvo[0] = (
            gc[1] * ((pf[0] + pf[3]) * sr1 + fvo[5] * sr2) + gc[0] * fmass * sl1
        ) * d
        fvo[1] = (
            gc[1] * (dconjg(fvo[5]) * sr1 + (pf[0] - pf[3]) * sr2) + gc[0] * fmass * sl2
        ) * d
        fvo[2] = (
            gc[0] * ((pf[0] - pf[3]) * sl1 - fvo[5] * sl2) + gc[1] * fmass * sr1
        ) * d
        fvo[3] = (
            gc[0] * (-dconjg(fvo[5]) * sl1 + (pf[0] + pf[3]) * sl2)
            + gc[1] * fmass * sr2
        ) * d
    else:
        d = d * gc[0]
        fvo[0] = fmass * sl1 * d
        fvo[1] = fmass * sl2 * d
        fvo[2] = ((pf[0] - pf[3]) * sl1 - fvo[5] * sl2) * d
        fvo[3] = (-dconjg(fvo[5]) * sl1 + (pf[0] + pf[3]) * sl2) * d

    return


#
# This subroutine computes an off-shell fermion wavefunction from a
# flowing-IN external fermion and a vector boson.
#
# input:
#       complex fi[5]          : flow-in  fermion                   |fi>
#       complex vc[5]          : input    vector                      v
#       complex gc[1]          : coupling constants                  gvf
#       real    fmass          : mass  of output fermion f'
#       real    fwidth         : width of output fermion f'
#
# output:
#       complex fvi[5]         : off-shell fermion             |f',v,fi>
#
# double complex fi[5],vc[5],gc[1],fvi[5],sl1,sl2,sr1,sr2,d
# double precision pf(0:3),fmass,fwidth,pf2
#
# double precision 0.0, 1.0
# parameter( 0.0 = 0.0, 1.0 = 1.0 )
# double complex cImag, cZero
# parameter( cImag = ( 0.0, 1.0 ), cZero = ( 0.0, 0.0 ) )
def fvixxx(fi, vc, gc, fmass, fwidth, fvi):
    fvi[4] = fi[4] - vc[4]
    fvi[5] = fi[5] - vc[5]
    pf[0] = dble(fvi[4])
    pf[1] = dble(fvi[5])
    pf[2] = dimag(fvi[5])
    pf[3] = dimag(fvi[4])
    pf2 = pf[0] ** 2 - (pf[1] ** 2 + pf[2] ** 2 + pf[3] ** 2)
    d = -1.0 / dcmplx(pf2 - fmass**2 + 1j * fmass * fwidth)
    sl1 = (vc[0] + vc[3]) * fi[0] + (vc[1] - cImag * vc[2]) * fi[1]
    sl2 = (vc[1] + cImag * vc[2]) * fi[0] + (vc[0] - vc[3]) * fi[1]

    if gc[1] != cZero:
        sr1 = (vc[0] - vc[3]) * fi[2] - (vc[1] - cImag * vc[2]) * fi[3]
        sr2 = -(vc[1] + cImag * vc[2]) * fi[2] + (vc[0] + vc[3]) * fi[3]
        fvi[0] = (
            gc[0] * ((pf[0] - pf[3]) * sl1 - dconjg(fvi[5]) * sl2) + gc[1] * fmass * sr1
        ) * d
        fvi[1] = (
            gc[0] * (-fvi[5] * sl1 + (pf[0] + pf[3]) * sl2) + gc[1] * fmass * sr2
        ) * d
        fvi[2] = (
            gc[1] * ((pf[0] + pf[3]) * sr1 + dconjg(fvi[5]) * sr2) + gc[0] * fmass * sl1
        ) * d
        fvi[3] = (
            gc[1] * (fvi[5] * sr1 + (pf[0] - pf[3]) * sr2) + gc[0] * fmass * sl2
        ) * d
    else:
        d = d * gc[0]
        fvi[0] = ((pf[0] - pf[3]) * sl1 - dconjg(fvi[5]) * sl2) * d
        fvi[1] = (-fvi[5] * sl1 + (pf[0] + pf[3]) * sl2) * d
        fvi[2] = fmass * sl1 * d
        fvi[3] = fmass * sl2 * d
    return


#
# This subroutine computes an off-shell fermion wavefunction from a
# flowing-IN external fermion and a vector boson.
#
# input:
#       complex fi[5]          : flow-in  fermion                   |fi>
#       complex sc[2]          : input    scalar                      s
#       complex gc[1]          : coupling constants                 gchf
#       real    fmass          : mass  of OUTPUT fermion f'
#       real    fwidth         : width of OUTPUT fermion f'
#
# output:
#       complex fsi[5]         : off-shell fermion             |f',s,fi>
#
# double complex fi[5],sc[2],fsi[5],gc[1],sl1,sl2,sr1,sr2,ds
# double precision pf(0:3),fmass,fwidth,pf2,p0p3,p0m3
def fsixxx(fi, sc, gc, fmass, fwidth, fsi):
    fsi[4] = fi[4] - sc[1]
    fsi[5] = fi[5] - sc[2]
    pf[0] = dble(fsi[4])
    pf[1] = dble(fsi[5])
    pf[2] = dimag(fsi[5])
    pf[3] = dimag(fsi[4])
    pf2 = pf[0] ** 2 - (pf[1] ** 2 + pf[2] ** 2 + pf[3] ** 2)
    ds = -sc[0] / (pf2 - fmass**2 + 1j * fmass * fwidth)
    p0p3 = pf[0] + pf[3]
    p0m3 = pf[0] - pf[3]
    sl1 = gc[0] * (p0p3 * fi[0] + dconjg(fsi[5]) * fi[1])
    sl2 = gc[0] * (p0m3 * fi[1] + fsi[5] * fi[0])
    sr1 = gc[1] * (p0m3 * fi[2] - dconjg(fsi[5]) * fi[3])
    sr2 = gc[1] * (p0p3 * fi[3] - fsi[5] * fi[2])
    fsi[0] = (gc[0] * fmass * fi[0] + sr1) * ds
    fsi[1] = (gc[0] * fmass * fi[1] + sr2) * ds
    fsi[2] = (gc[1] * fmass * fi[2] + sl1) * ds
    fsi[3] = (gc[1] * fmass * fi[3] + sl2) * ds
    return


#
# This subroutine computes an off-shell fermion wavefunction from a
# flowing-OUT external fermion and a vector boson.
#
# input:
#       complex fo[5]          : flow-out fermion                   <fo|
#       complex sc[5]          : input    scalar                      s
#       complex gc[1]          : coupling constants                 gchf
#       real    fmass          : mass  of OUTPUT fermion f'
#       real    fwidth         : width of OUTPUT fermion f'
#
# output:
#       complex fso[5]         : off-shell fermion             <fo,s,f'|
#

# double complex fo[5],sc[5],fso[5],gc[1],sl1,sl2,sr1,sr2,ds
# double precision pf(0:3),fmass,fwidth,pf2,p0p3,p0m3
def fsoxxx(fo, sc, gc, fmass, fwidth, fso):
    fso[4] = fo[4] + sc[1]
    fso[5] = fo[5] + sc[2]
    pf[0] = dble(fso[4])
    pf[1] = dble(fso[5])
    pf[2] = dimag(fso[5])
    pf[3] = dimag(fso[4])
    pf2 = pf[0] ** 2 - (pf[1] ** 2 + pf[2] ** 2 + pf[3] ** 2)
    ds = -sc[0] / dcmplx(pf2 - fmass**2, fmass * fwidth)
    p0p3 = pf[0] + pf[3]
    p0m3 = pf[0] - pf[3]
    sl1 = gc[1] * (p0p3 * fo[2] + fso[5] * fo[3])
    sl2 = gc[1] * (p0m3 * fo[3] + dconjg(fso[5]) * fo[2])
    sr1 = gc[0] * (p0m3 * fo[0] - fso[5] * fo[1])
    sr2 = gc[0] * (p0p3 * fo[1] - dconjg(fso[5]) * fo[0])
    fso[0] = (gc[0] * fmass * fo[0] + sl1) * ds
    fso[1] = (gc[0] * fmass * fo[1] + sl2) * ds
    fso[2] = (gc[1] * fmass * fo[2] + sr1) * ds
    fso[3] = (gc[1] * fmass * fo[3] + sr2) * ds
    return


#
# This subroutine computes an amplitude of the fermion-fermion-scalar
# coupling.
#
# input:
#       complex fi[5]          : flow-in  fermion                   |fi>
#       complex fo[5]          : flow-out fermion                   <fo|
#       complex sc[2]          : input    scalar                      s
#       complex gc[1]          : coupling constants                 gchf
#
# output:
#       complex vertex         : amplitude                     <fo|s|fi>
#
# double complex fi[5],fo[5],sc[2],gc[1],vertex
# INTEGER DIM
# PARAMETER(DIM=18)
def iosxxx(fi, fo, sc, gc, vertex):
    vertex = sc[0] * (
        gc[0] * (fi[0] * fo[0] + fi[1] * fo[1])
        + gc[1] * (fi[2] * fo[2] + fi[3] * fo[3])
    )
    return



from helax.numpy.typing import RealArray


def check_momentum_shape(momentum: RealArray, require_batch: bool = True):
    shape = momentum.shape
    assert shape[0] == 4, "Leading dimension must have length 4."
    if require_batch:
        assert len(shape) == 2, "Expected a batch dimension. Found 1-D array."



from typing import NamedTuple, Union

RealOrComplex = Union[float, complex]


# =============================================================================
# ---- Renormalizable Vertices ------------------------------------------------
# =============================================================================


class VertexFFS(NamedTuple):
    """
    Vertex representing fermion-fermion-scalar vertex corresponding to the
    interaction: S * fbar * (gL * PL + gR * PR) * f

    Parameters
    ----------
    left: float
        Coupling coefficient of PL
    right: float
        Coupling coefficient of PR
    """

    left: RealOrComplex
    right: RealOrComplex


class VertexFFV(NamedTuple):
    """
    Vertex representing fermion-fermion-vector vertex corresponding to the
    interaction: V[mu] * fbar * gamma[mu] * (gL * PL + gR * PR) * f

    Parameters
    ----------
    left: float
        Coupling coefficient of gamma[mu] * PL
    right: float
        Coupling coefficient of gamma[mu] * PR
    """

    left: RealOrComplex
    right: RealOrComplex


class VertexWWS(NamedTuple):
    """
    Vertex representing weyl-weyl-scalar vertex corresponding to the
    interaction: g * S * chi * chi

    Parameters
    ----------
    g: complex
        Coupling coefficient.
    """

    g: RealOrComplex


class VertexWWV(NamedTuple):
    """
    Vertex representing weyl-weyl-vector vertex corresponding to the
    interaction: g * V[mu] * chi^+ * SigmaBar[mu] * eta

    Parameters
    ----------
    g: float
        Coupling coefficient.
    """

    g: RealOrComplex


class VertexVVVV(NamedTuple):
    """
    Vertex representing four-vector vertex corresponding to the
    interaction: V1[mu1] V2[mu2] V3[mu3] V4[mu4]

    Parameters
    ----------
    g1,g2,g3: float or complex
        Coupling coefficients of:
            g1: g[mu1,mu2] * g[mu3, mu4]
            g2: g[mu1,mu3] * g[mu2, mu4]
            g3: g[mu1,mu4] * g[mu2, mu3]
    """

    g1: RealOrComplex
    g2: RealOrComplex
    g3: RealOrComplex


class VertexVVV(NamedTuple):
    """
    Vertex representing three-vector vertex.

    The interaction term is:
          g[mu,nu] V1[mu] V2[nu] del[V3[rho], rho]
        + g[mu,nu] V2[mu] V3[nu] del[V1[rho], rho]
        + g[mu,nu] V3[mu] V1[nu] del[V2[rho], rho]
    with the vertex rule:
        g * (
              g[mu,nu] * (k2[rho] - k1[rho])
            + g[nu,rho] * (k3[mu] - k2[mu])
            + g[rho,mu] * (k1[nu] - k3[nu])
        )

    Parameters
    ----------
    g: float or complex
        Coupling coefficient
    """

    g: RealOrComplex


class VertexSSSS(NamedTuple):
    """
    Vertex representing a four-scalar vertex.
    """

    g: RealOrComplex


class VertexSSS(NamedTuple):
    """
    Vertex representing a four-scalar vertex.
    """

    g: RealOrComplex


class VertexSSVV(NamedTuple):
    """
    Vertex representing a scalar-scalar-vector-vector vertex.
    """

    g: RealOrComplex


class VertexSSV(NamedTuple):
    """
    Vertex representing a scalar-scalar-vector vertex.
    """

    g: RealOrComplex


class VertexSVV(NamedTuple):
    """
    Vertex representing a scalar-vector-vector vertex.
    """

    g: RealOrComplex


class VertexUUV(NamedTuple):
    """
    Vertex representing a ghost-ghost-vector vertex.
    """

    g1: RealOrComplex
    g2: RealOrComplex


class VertexUUS(NamedTuple):
    """
    Vertex representing a ghost-ghost-scalar vertex.
    """

    g: RealOrComplex


# =============================================================================
# ---- Effective Vertices -----------------------------------------------------
# =============================================================================


class VertexFFSDeriv(NamedTuple):
    """
    Vertex representing fermion-fermion-scalar vertex corresponding to the
    interaction: d[S,mu] * fbar * gamma[mu] * (gL * PL + gR * PR) * f

    Parameters
    ----------
    left: float
        Coupling coefficient of gamma[mu] * PL
    right: float
        Coupling coefficient of gamma[mu] * PR
    """

    left: RealOrComplex
    right: RealOrComplex



import dataclasses

import numpy as np

from helax.numpy.typing import ComplexArray, RealArray


@dataclasses.dataclass
class DiracWf:
    """Dataclass for Dirac wavefunctions.

    Attributes
    ----------
    wavefunction: complex array
        4-Component Dirac spinor.
    momentum: array
        Four-momentum of the spinor. The momentum always points in the
        direction of the fermion flow.
    direction: array
        1 for a u,v spinors and -1 for ubar,vbar spinors.

    """

    wavefunction: ComplexArray
    momentum: RealArray
    direction: int


# =============================================================================
# ---- Dirac Wavefunction -----------------------------------------------------
# =============================================================================


def __check_spin_dirac(*, spin: int):
    assert spin == 1 or spin == -1, "Spin must be 1 or -1."


def __chi(*, momentum: RealArray, spin: int) -> ComplexArray:
    """
    Compute the two-component weyl spinor.

    Parameters
    ----------
    p: array
        Array containing the 4-momentum of the wavefunction.
    s: int
        Spin of the wavefunction. Must be 1 or -1.
    """
    dtype = momentum.dtype
    eps = np.finfo(dtype.name).eps

    px = momentum[1]
    py = momentum[2]
    pz = momentum[3]

    pm = np.linalg.norm(momentum[1:], axis=0)
    mask = pm + pz <= eps
    x = np.zeros((2, momentum.shape[-1]), dtype=dtype) * 1j

    if np.any(mask):
        # x[0, mask] = 0.0j  # (spin - 1.0) / 2.0 + 0j
        x[1, mask] = spin  # (spin + 1.0) / 2.0 + 0j

    mask = ~mask
    if np.any(mask):
        px = px[mask]
        py = py[mask]
        pz = pz[mask]
        pm = pm[mask]

        den = np.sqrt(2 * pm * (pm + pz))
        x[0, mask] = (pm + pz) / den
        x[1, mask] = (spin * px + py * 1j) / den

    if spin == -1:
        return np.array([x[1], x[0]])
    return np.array([x[0], x[1]])


def __dirac_spinor(
    *, momentum: RealArray, mass: float, spin: int, anti: int
) -> ComplexArray:
    """
    Compute the dirac wavefunction.

    Parameters
    ----------
    p:
        Four-momentum of the wavefunction.
    mass:
        Mass of the wavefunction.
    s: int
        Spin of the wavefunction. Must be 1 or -1.
    anti: Int
        If anti = -1, the wavefunction represents a v-spinor.
        If 1, wavefunction represents a u-spinor.
    """
    pm = np.linalg.norm(momentum[1:], axis=0)

    wp = np.sqrt(momentum[0] + pm)
    wm = mass / wp

    if spin == -anti:
        w = np.array([anti * wp, wm])
    else:
        w = np.array([wm, anti * wp])

    x = __chi(momentum=momentum, spin=spin * anti)

    return np.array([w[0] * x[0], w[0] * x[1], w[1] * x[0], w[1] * x[1]])


def __spinor_u(*, momentum: RealArray, mass: float, spin: int) -> ComplexArray:
    return __dirac_spinor(momentum=momentum, mass=mass, spin=spin, anti=1)


def __spinor_v(*, momentum: RealArray, mass: float, spin: int) -> ComplexArray:
    return __dirac_spinor(momentum=momentum, mass=mass, spin=spin, anti=-1)


def __spinor_ubar(*, momentum: RealArray, mass: float, spin: int) -> ComplexArray:
    x = np.conj(__dirac_spinor(momentum=momentum, mass=mass, spin=spin, anti=1))
    return np.array([x[2], x[3], x[0], x[1]])


def __spinor_vbar(*, momentum: RealArray, mass: float, spin: int) -> ComplexArray:
    x = np.conj(__dirac_spinor(momentum=momentum, mass=mass, spin=spin, anti=-1))
    return np.array([x[2], x[3], x[0], x[1]])


def spinor_u(momentum: RealArray, mass: float, spin: int) -> DiracWf:
    """
    Compute a u-spinor wavefunction.

    Parameters
    ----------
    momentum: ndarray
        Array containing the four-momentum of the particle.
        Must be 1 or 2 dimensional with leading dimension of size 4.
        If 2-dimensional, 2nd dimension must be the batch dimension.
    mass: float
        Mass of the particle.
    spin: int
        Spin of the particle. Must be 1 or -1.
    """
    __check_spin_dirac(spin=spin)
    wf = __spinor_u(momentum=momentum, mass=mass, spin=spin)
    return DiracWf(
        wavefunction=wf,
        momentum=momentum,
        direction=1,
    )


def spinor_v(momentum: RealArray, mass: float, spin: int) -> DiracWf:
    """
    Compute a v-spinor wavefunction.

    Parameters
    ----------
    momentum: ndarray
        Array containing the four-momentum of the particle.
        Must be 1 or 2 dimensional with leading dimension of size 4.
    mass: float
        Mass of the particle.
    spin: int
        Spin of the particle. Must be 1 or -1.
    """
    __check_spin_dirac(spin=spin)
    wf = __spinor_v(momentum=momentum, mass=mass, spin=spin)
    return DiracWf(
        wavefunction=wf,
        momentum=momentum,
        direction=1,
    )


def spinor_ubar(momentum: RealArray, mass: float, spin: int) -> DiracWf:
    """
    Compute a ubar-spinor wavefunction.

    Parameters
    ----------
    momentum: ndarray
        Array containing the four-momentum of the particle.
        Must be 1 or 2 dimensional with leading dimension of size 4.
    mass: float
        Mass of the particle.
    spin: int
        Spin of the particle. Must be 1 or -1.
    """
    __check_spin_dirac(spin=spin)
    wf = __spinor_ubar(momentum=momentum, mass=mass, spin=spin)
    return DiracWf(
        wavefunction=wf,
        momentum=-momentum,
        direction=-1,
    )


def spinor_vbar(momentum: RealArray, mass: float, spin: int) -> DiracWf:
    """
    Compute a vbar-spinor wavefunction.

    Parameters
    ----------
    momentum: ndarray
        Array containing the four-momentum of the particle.
        Must be 1 or 2 dimensional with leading dimension of size 4.
    mass: float
        Mass of the particle.
    spin: int
        Spin of the particle. Must be 1 or -1.
    """
    __check_spin_dirac(spin=spin)
    wf = __spinor_vbar(momentum=momentum, mass=mass, spin=spin)
    return DiracWf(
        wavefunction=wf,
        momentum=-momentum,
        direction=-1,
    )


def charge_conjugate(psi: DiracWf) -> DiracWf:
    """
    Charge conjugate the input wavefunction.

    Parameters
    ----------
    psi: DiracWf
        The dirac wavefunction.

    Returns
    -------
    psi_cc: DiracWf
        Charge conjugated wavefunction.
    """
    s = psi.direction
    wf = np.array(
        [
            s * psi.wavefunction[1],
            -s * psi.wavefunction[0],
            -s * psi.wavefunction[3],
            s * psi.wavefunction[2],
        ]
    )
    p = -psi.momentum
    return DiracWf(wavefunction=wf, momentum=p, direction=-s)



try:
    import jax
    import jax.numpy

    from . import amplitudes, lvector, phase_space, utils, wavefunctions

    __all__ = [
        "amplitudes",
        "lvector",
        "phase_space",
        "utils",
        "wavefunctions",
    ]

except ImportError as e:
    print("`jax` is required to use `helax.jax`.")
    raise e



import pathlib
import unittest
from typing import Callable

import numpy as np
import pytest

from helax.numpy.dirac import Dirac1, DiracG0, DiracG1, DiracG2, DiracG3
from helax.numpy.wavefunctions import (DiracWf, spinor_u, spinor_ubar,
                                       spinor_v, spinor_vbar)

TEST_DATA = np.load(
    pathlib.Path(__file__).parent.joinpath("testdata").joinpath("spinor_data.npz")
)


def _sigma_momentum(momentum):
    mass = np.sqrt(
        momentum[0] ** 2 - momentum[1] ** 2 - momentum[2] ** 2 - momentum[3] ** 2
    )
    return (
        momentum[0] * DiracG0
        - momentum[1] * DiracG1
        - momentum[2] * DiracG2
        - momentum[3] * DiracG3
        + mass * Dirac1
    )


class TestDiracCompleteness(unittest.TestCase):
    def setUp(self) -> None:
        self.mass = 4.0
        self.momenta = np.transpose(
            np.array(
                [
                    [5.0, 0.0, 0.0, 3.0],
                    [5.0, 0.0, 0.0, -3.0],
                ]
            )
        )

    def test_completeness_spinor_u(self):
        # shapes: (num_spins, 2, num_momenta)
        wf_u = np.squeeze(
            np.array(
                [spinor_u(self.momenta, self.mass, s).wavefunction for s in (-1, 1)]
            )
        )
        wf_ubar = np.squeeze(
            np.array(
                [spinor_ubar(self.momenta, self.mass, s).wavefunction for s in (-1, 1)]
            )
        )

        for i in range(self.momenta.shape[-1]):
            # shapes: (num_spins, 4)
            u = wf_u[..., i]
            ubar = wf_ubar[..., i]

            spin_sum = np.einsum("ij,ik", u, ubar)
            expect = _sigma_momentum(self.momenta[..., i])
            self.assertLess(np.max(np.abs(spin_sum - expect)), 1e-10)


def run_spinor_tests(
    fn: Callable[[np.ndarray, float, int], DiracWf], ty: str, massive: bool
):
    assert ty in ["u", "v", "ubar", "vbar"], "Invalid string passed to test runner."
    if massive:
        prefix = ty + "_massive_"
        mass = 3.0
    else:
        prefix = ty + "_massless_"
        mass = 0.0

    momenta = TEST_DATA[prefix + "momenta"]
    spin_up: np.ndarray = TEST_DATA[prefix + "up"]
    spin_down: np.ndarray = TEST_DATA[prefix + "down"]

    helax_spin_up: np.ndarray = np.transpose(fn(momenta.T, mass, 1).wavefunction)
    helax_spin_down = np.transpose(fn(momenta.T, mass, -1).wavefunction)

    for tu, td, hu, hd in zip(spin_up, spin_down, helax_spin_up, helax_spin_down):
        for i in range(4):
            assert np.real(hu[i]) == pytest.approx(np.real(tu[i]), rel=1e-4, abs=0.0)
            assert np.real(hd[i]) == pytest.approx(np.real(td[i]), rel=1e-4, abs=0.0)
            assert np.imag(hu[i]) == pytest.approx(np.imag(tu[i]), rel=1e-4, abs=0.0)
            assert np.imag(hd[i]) == pytest.approx(np.imag(td[i]), rel=1e-4, abs=0.0)

    # Special case: pm == -pz
    if massive:
        e = 2.0
        mass = np.sqrt(3)
    else:
        e = 1.0
        mass = 0.0

    p = np.expand_dims(np.array([e, 0.0, 0.0, -1.0]), -1)
    em = np.sqrt(e - 1)
    ep = np.sqrt(e + 1)

    if ty == "u":
        spin_up = np.array([0, em, 0, ep])
        spin_down = np.array([-ep, 0, -em, 0])
    elif ty == "v":
        spin_up = np.array([ep, 0, -em, 0])
        spin_down = np.array([0, em, 0, -ep])
    elif ty == "ubar":
        spin_up = np.array([0, ep, 0, em])
        spin_down = np.array([-em, 0, -ep, 0])
    else:
        spin_up = np.array([-em, 0, ep, 0])
        spin_down = np.array([0, -ep, 0, em])

    helax_spin_up = fn(p, mass, 1).wavefunction
    helax_spin_down = fn(p, mass, -1).wavefunction

    assert np.real(helax_spin_up[0, 0]) == pytest.approx(spin_up[0])
    assert np.real(helax_spin_up[1, 0]) == pytest.approx(spin_up[1])
    assert np.real(helax_spin_up[2, 0]) == pytest.approx(spin_up[2])
    assert np.real(helax_spin_up[3, 0]) == pytest.approx(spin_up[3])

    assert np.real(helax_spin_down[0, 0]) == pytest.approx(spin_down[0])
    assert np.real(helax_spin_down[1, 0]) == pytest.approx(spin_down[1])
    assert np.real(helax_spin_down[2, 0]) == pytest.approx(spin_down[2])
    assert np.real(helax_spin_down[3, 0]) == pytest.approx(spin_down[3])


def test_spinor_u_massive():
    run_spinor_tests(spinor_u, "u", massive=True)


def test_spinor_v_massive():
    run_spinor_tests(spinor_v, "v", massive=True)


def test_spinor_ubar_massive():
    run_spinor_tests(spinor_ubar, "ubar", massive=True)


def test_spinor_vbar_massive():
    run_spinor_tests(spinor_vbar, "vbar", massive=True)


def test_spinor_u_massless():
    run_spinor_tests(spinor_u, "u", massive=False)


def test_spinor_v_massless():
    run_spinor_tests(spinor_v, "v", massive=False)


def test_spinor_ubar_massless():
    run_spinor_tests(spinor_ubar, "ubar", massive=False)


def test_spinor_vbar_massless():
    run_spinor_tests(spinor_vbar, "vbar", massive=False)



import functools
from typing import Tuple

import jax
import jax.numpy as jnp
import numpy as np
import pytest

from helax.jax import amplitudes, phase_space, wavefunctions
from helax.jax.lvector import lnorm_sqr
from helax.jax.utils import abs2, kallen_lambda
from helax.jax.wavefunctions import DiracWf, ScalarWf, VectorWf

MUON_MASS = 1.056584e-01
G_FERMI = 1.166379e-05
EL = 3.028620e-01
SW = 4.808530e-01
MASS_W = 8.038500e01
WIDTH_W = 2.085000e00
MASS_B = 4.18
MASS_T = 172.9
MASS_MU = 105.6583745e-3
MASS_E = 0.5109989461e-3

MASS_H = 125.00
WIDTH_H = 0.00374
VEV_H = 246.21965


@jax.jit
def spinor_u(momenta, mass: float) -> Tuple[DiracWf, DiracWf]:
    return (
        wavefunctions.spinor_u(momenta, mass, -1),
        wavefunctions.spinor_u(momenta, mass, 1),
    )


@jax.jit
def spinor_v(momenta, mass: float) -> Tuple[DiracWf, DiracWf]:
    return (
        wavefunctions.spinor_v(momenta, mass, -1),
        wavefunctions.spinor_v(momenta, mass, 1),
    )


@jax.jit
def spinor_ubar(momenta, mass: float) -> Tuple[DiracWf, DiracWf]:
    return (
        wavefunctions.spinor_ubar(momenta, mass, -1),
        wavefunctions.spinor_ubar(momenta, mass, 1),
    )


@jax.jit
def spinor_vbar(momenta, mass: float) -> Tuple[DiracWf, DiracWf]:
    return (
        wavefunctions.spinor_vbar(momenta, mass, -1),
        wavefunctions.spinor_vbar(momenta, mass, 1),
    )


def polvec(momenta, mass: float, out: bool) -> Tuple[VectorWf, VectorWf, VectorWf]:
    return (
        wavefunctions.vector_wf(momenta, mass, -1, out),
        wavefunctions.vector_wf(momenta, mass, 0, out),
        wavefunctions.vector_wf(momenta, mass, 1, out),
    )


def scalar_wf(momenta, out: bool) -> ScalarWf:
    return wavefunctions.scalar_wf(momenta, out)


@functools.partial(jax.jit, static_argnums=(0, 1, 2))
def current_ff_to_v(
    vertex, mass: float, width: float, psi_out: DiracWf, psi_in: DiracWf
):
    return amplitudes.current_ff_to_v(vertex, mass, width, psi_out, psi_in)


@functools.partial(jax.jit, static_argnums=(0, 1, 2))
def current_ff_to_s(vertex, mass, width, psi_out: DiracWf, psi_in: DiracWf):
    return amplitudes.current_ff_to_s(vertex, mass, width, psi_out, psi_in)


@functools.partial(jax.jit, static_argnums=(0,))
def amplitude_ffv(vertex, psi_out: DiracWf, psi_in: DiracWf, eps: VectorWf):
    return amplitudes.amplitude_ffv(vertex, psi_out, psi_in, eps)


@functools.partial(jax.jit, static_argnums=(0,))
def amplitude_ffs(vertex, psi_out: DiracWf, psi_in: DiracWf, phi: ScalarWf):
    return amplitudes.amplitude_ffs(vertex, psi_out, psi_in, phi)


def wavefunction_iter(*wavefuncs):
    idx_list = []
    for wf in wavefuncs:
        if isinstance(wf, wavefunctions.ScalarWf):
            idx_list.append([0])
        elif isinstance(wf, wavefunctions.DiracWf):
            idx_list.append([0, 1])
        elif isinstance(wf, wavefunctions.VectorWf):
            idx_list.append([0, 1, 2])
        else:
            raise ValueError(f"Invalid wf {type(wf)}")

    idxs = np.array(np.meshgrid(*idx_list)).T.reshape(-1, len(idx_list))
    for idx in idxs:
        yield tuple([wf[i] for wf, i in zip(wavefuncs, idx)])


def test_width_mu_to_e_nu_nu():
    def msqrd_mu_to_e_nu_nu_analytic(momenta):
        t = lnorm_sqr(momenta[:, 0] + momenta[:, 2])
        s = lnorm_sqr(momenta[:, 1] + momenta[:, 1])

        return (
            -0.125
            * (
                EL**4
                * (
                    MASS_E**4 * MASS_MU**2 * (MASS_MU**2 - s - t)
                    + 4 * MASS_W**4 * t * (-(MASS_MU**2) + t)
                    + MASS_E**2
                    * (
                        -4 * MASS_W**4 * t
                        - MASS_MU**4 * (s + t)
                        + MASS_MU**2
                        * (4 * MASS_W**4 + 4 * MASS_W**2 * s + (s + t) ** 2)
                    )
                )
            )
            / (
                MASS_W**4
                * SW**4
                * (
                    (-(MASS_E**2) - MASS_MU**2 + MASS_W**2 + s + t) ** 2
                    + MASS_W**2 * WIDTH_W**2
                )
            )
        )

    def msqrd_mu_to_e_nu_nu_hel(momenta):
        pe = momenta[:, 0]
        pve = momenta[:, 1]
        pvm = momenta[:, 2]
        pmu = jnp.sum(momenta, axis=1)

        v_wll = amplitudes.VertexFFV(left=EL / (np.sqrt(2) * SW), right=0.0)

        mu_wfs = spinor_u(pmu, MUON_MASS)
        e_wfs = spinor_ubar(pe, 0.0)
        ve_wfs = spinor_v(pve, 0.0)
        vm_wfs = spinor_ubar(pvm, 0.0)

        def msqrd(imu, ie, ive, ivm):
            w_wf = current_ff_to_v(v_wll, MASS_W, WIDTH_W, vm_wfs[ivm], mu_wfs[imu])
            amp = amplitude_ffv(v_wll, e_wfs[ie], ve_wfs[ive], w_wf)
            return abs2(amp)

        idxs = np.array(np.meshgrid([0, 1], [0, 1], [0, 1], [0, 1])).T.reshape(-1, 4)
        res = sum(msqrd(i_n, i_v1, i_v2, i_v3) for (i_n, i_v1, i_v2, i_v3) in idxs)

        return res / 2.0

    key = jax.random.PRNGKey(1234)
    width = phase_space.decay_width(
        MASS_MU, np.array([MASS_E, 0.0, 0.0]), key, msqrd=msqrd_mu_to_e_nu_nu_hel
    )[0]

    width2 = phase_space.decay_width(
        MASS_MU, np.array([MASS_E, 0.0, 0.0]), key, msqrd=msqrd_mu_to_e_nu_nu_analytic
    )[0]

    assert float(width) == pytest.approx(width2, abs=0.0, rel=0.01)


def test_width_t_to_b_w():
    def msqrd_t_to_b_w(momenta):
        pb = momenta[:, 0]
        pw = momenta[:, 1]
        pt = jnp.sum(momenta, axis=1)

        t_wfs = spinor_u(pt, MASS_T)
        b_wfs = spinor_ubar(pb, MASS_B)
        w_wfs = polvec(pw, MASS_W, True)

        v_tbw = amplitudes.VertexFFV(left=EL / (np.sqrt(2.0) * SW), right=0.0)

        def msqrd(it, ib, iw):
            amp = amplitude_ffv(v_tbw, b_wfs[ib], t_wfs[it], w_wfs[iw])
            return abs2(amp)

        idxs = np.array(np.meshgrid([0, 1], [0, 1], [0, 1, 2])).T.reshape(-1, 3)
        res = sum(msqrd(it, ib, iw) for (it, ib, iw) in idxs)

        return res / 2.0

    analytic = (
        EL**2
        * (
            MASS_B**4
            + MASS_T**4
            + MASS_T**2 * MASS_W**2
            - 2 * MASS_W**4
            + MASS_B**2 * (-2 * MASS_T**2 + MASS_W**2)
        )
        * np.sqrt(kallen_lambda(MASS_B**2, MASS_T**2, MASS_W**2))
    ) / (64 * MASS_T**3 * MASS_W**2 * np.pi * SW**2)

    key = jax.random.PRNGKey(1234)
    width = phase_space.decay_width(
        MASS_T, np.array([MASS_B, MASS_W]), key, msqrd=msqrd_t_to_b_w
    )[0]

    assert float(width) == pytest.approx(analytic, rel=1e-3, abs=0.0)


def test_width_h_to_b_b():
    def msqrd_h_to_b_b(momenta):
        p1 = momenta[:, 0]
        p2 = momenta[:, 1]
        ph = jnp.sum(momenta, axis=1)

        wf1s = spinor_ubar(p1, MASS_B)
        wf2s = spinor_v(p2, MASS_B)
        h_wf = scalar_wf(ph, False)

        v_hbb = amplitudes.VertexFFS(
            left=EL * MASS_B / (2 * MASS_W * SW), right=EL * MASS_B / (2 * MASS_W * SW)
        )

        def msqrd(i1, i2):
            return abs2(amplitude_ffs(v_hbb, wf1s[i2], wf2s[i1], h_wf))

        idxs = np.array(np.meshgrid([0, 1], [0, 1])).T.reshape(-1, 2)
        res = sum(msqrd(i1, i2) for (i1, i2) in idxs)

        return res

    mub = MASS_B / MASS_H
    analytic = (
        EL**2
        * MASS_H**3
        * mub**2
        * (1 - 4 * mub**2) ** 1.5
        / (32 * MASS_W**2 * np.pi * SW**2)
    )

    key = jax.random.PRNGKey(1234)
    width = phase_space.decay_width(
        MASS_H, np.array([MASS_B, MASS_B]), key, msqrd=msqrd_h_to_b_b
    )[0]

    assert float(width) == pytest.approx(analytic, rel=1e-3, abs=0.0)


def test_dark_matter_annihilation_scalar_mediator():
    gsxx = 1.0
    gsff = 1.0
    mx = 10.0
    mf = 1.0
    cme = 3 * mx
    mass_s = 1.0
    width_s = 1.0

    vxx = amplitudes.VertexFFS(left=gsxx, right=gsxx)
    vff = amplitudes.VertexFFS(left=gsff, right=gsff)

    ex = cme / 2.0
    px = np.sqrt(ex**2 - mx**2)
    p1 = jnp.expand_dims(jnp.array([ex, 0.0, 0.0, px]), -1)
    p2 = jnp.expand_dims(jnp.array([ex, 0.0, 0.0, -px]), -1)

    wf1s = spinor_u(p1, mx)
    wf2s = spinor_vbar(p2, mx)

    def msqrd_x_x_to_f_f(momenta):
        p3 = momenta[:, 0]
        p4 = momenta[:, 1]

        wf3s = spinor_ubar(p3, mf)
        wf4s = spinor_v(p4, mf)

        def msqrd(i1, i2, i3, i4):
            swf = current_ff_to_s(vxx, mass_s, width_s, wf2s[i2], wf1s[i1])
            return abs2(amplitude_ffs(vff, wf3s[i3], wf4s[i4], swf))

        idxs = np.array(np.meshgrid([0, 1], [0, 1], [0, 1], [0, 1])).T.reshape(-1, 4)
        res = sum(msqrd(i1, i2, i3, i4) for (i1, i2, i3, i4) in idxs)

        return res / 4.0

    key = jax.random.PRNGKey(4)
    cs = phase_space.cross_section(
        cme, mx, mx, np.array([mf, mf]), key, msqrd=msqrd_x_x_to_f_f
    )[0]

    s = cme**2
    analytic = (
        gsff**2
        * gsxx**2
        * (s - 4 * mf**2)
        * np.sqrt((s - 4 * mf**2) * (s - 4 * mx**2))
    ) / (16 * np.pi * s * ((s - mass_s**2) ** 2 + (mass_s * width_s) ** 2))

    assert float(cs) == pytest.approx(analytic, rel=1e-2, abs=0.0)


def test_dark_matter_annihilation_vector_mediator():
    gvxx = 1.0
    gvff = 1.0
    mx = 10.0
    mf = 1.0
    cme = 3 * mx
    mv = 1.0
    widthv = 1.0

    vxx = amplitudes.VertexFFV(left=gvxx, right=gvxx)
    vff = amplitudes.VertexFFV(left=gvff, right=gvff)

    ex = cme / 2.0
    px = np.sqrt(ex**2 - mx**2)
    p1 = jnp.expand_dims(jnp.array([ex, 0.0, 0.0, px]), -1)
    p2 = jnp.expand_dims(jnp.array([ex, 0.0, 0.0, -px]), -1)

    wf1s = spinor_u(p1, mx)
    wf2s = spinor_vbar(p2, mx)

    def msqrd_x_x_to_f_f(momenta):
        p3 = momenta[:, 0]
        p4 = momenta[:, 1]

        wf3s = spinor_ubar(p3, mf)
        wf4s = spinor_v(p4, mf)

        def msqrd(i1, i2, i3, i4):
            vwf = current_ff_to_v(vxx, mv, widthv, wf2s[i2], wf1s[i1])
            return abs2(amplitude_ffv(vff, wf3s[i3], wf4s[i4], vwf))

        idxs = np.array(np.meshgrid([0, 1], [0, 1], [0, 1], [0, 1])).T.reshape(-1, 4)
        res = sum(msqrd(i1, i2, i3, i4) for (i1, i2, i3, i4) in idxs)

        return res / 4.0

    key = jax.random.PRNGKey(4)
    cs = phase_space.cross_section(
        cme, mx, mx, np.array([mf, mf]), key, msqrd=msqrd_x_x_to_f_f
    )[0]

    s = cme**2
    analytic = (
        gvff**2
        * gvxx**2
        * np.sqrt(s * (-4 * mf**2 + s))
        * (2 * mf**2 + s)
        * (2 * mx**2 + s)
    ) / (
        12.0
        * np.pi
        * s
        * np.sqrt(s * (-4 * mx**2 + s))
        * (mv**4 + s**2 + mv**2 * (-2 * s + widthv**2))
    )

    assert float(cs) == pytest.approx(analytic, rel=1e-2, abs=0.0)



import chex
import jax
import jax.numpy as jnp

from helax.vertices import VertexFFS, VertexFFV

from .lvector import ldot, lnorm_sqr
from .wavefunctions import DiracWf, ScalarWf, VectorWf

im = 1.0j


# =============================================================================
# ---- Propagators ------------------------------------------------------------
# =============================================================================


def complex_mass_sqr(mass: float, width: float) -> complex:
    return mass**2 - im * mass * width


def propagator_den(momentum: chex.Array, mass: float, width: float) -> chex.Array:
    return im / (lnorm_sqr(momentum) - complex_mass_sqr(mass, width))


def attach_dirac(psi: DiracWf, mass: float, width: float) -> DiracWf:
    p = psi.momentum
    f = psi.wavefunction

    den = propagator_den(psi.momentum, mass, width)

    p1p2 = p[1] + im * p[2]
    p1m2 = p[1] - im * p[2]
    p0p3 = p[0] + p[3]
    p0m3 = p[0] - p[3]

    wavefunction = jax.lax.switch(
        (psi.direction + 1) // 2,
        # (p_slash + m).psi
        (
            [
                (mass * f[0] - f[3] * p1m2 + f[2] * p0m3) * den,
                (mass * f[1] - f[2] * p1p2 + f[3] * p0p3) * den,
                (mass * f[2] + f[1] * p1m2 + f[0] * p0p3) * den,
                (mass * f[3] + f[0] * p1p2 + f[1] * p0m3) * den,
            ]
        ),
        # psi.(p_slash + m)
        (
            [
                (mass * f[0] + f[3] * p1p2 + f[2] * p0p3) * den,
                (mass * f[1] + f[2] * p1m2 + f[3] * p0m3) * den,
                (mass * f[2] - f[1] * p1p2 + f[0] * p0m3) * den,
                (mass * f[3] - f[0] * p1m2 + f[1] * p0p3) * den,
            ]
        ),
    )

    return DiracWf(
        wavefunction=wavefunction, momentum=psi.momentum, direction=psi.direction
    )


def attach_vector(eps: VectorWf, mass: float, width: float) -> VectorWf:
    wf = eps.wavefunction
    k = eps.momentum
    den = propagator_den(k, mass, width)

    invcm2 = jax.lax.cond(
        mass == 0.0, lambda: 0.0j, lambda: 1.0 / complex_mass_sqr(mass, width)
    )

    wf = (-wf + k * ldot(k, wf) * invcm2) * den

    return VectorWf(wavefunction=wf, momentum=k, direction=eps.direction)


def attach_scalar(phi: ScalarWf, mass: float, width: float) -> ScalarWf:
    wf = phi.wavefunction * propagator_den(phi.momentum, mass, width)
    return ScalarWf(wavefunction=wf, momentum=phi.momentum, direction=phi.direction)


# =============================================================================
# ---- Currents ---------------------------------------------------------------
# =============================================================================


def current_fs_to_f(
    vertex: VertexFFS, mass: float, width: float, psi: DiracWf, phi: ScalarWf
) -> DiracWf:
    phi_wf = phi.wavefunction
    fi = psi.wavefunction
    vl = vertex.left
    vr = vertex.right

    momentum = psi.momentum * psi.direction + phi.momentum

    wavefunction = jnp.array(
        [
            vl * phi_wf * fi[0],
            vl * phi_wf * fi[1],
            vr * phi_wf * fi[2],
            vr * phi_wf * fi[3],
        ]
    )

    psi = DiracWf(wavefunction=wavefunction, momentum=momentum, direction=psi.direction)
    return attach_dirac(psi, mass, width)


def current_ff_to_s(
    vertex: VertexFFS, mass: float, width: float, psi_out: DiracWf, psi_in: DiracWf
) -> ScalarWf:
    fi = psi_in.wavefunction
    fo = psi_out.wavefunction
    vl = vertex.left
    vr = vertex.right

    momentum = psi_in.momentum - psi_out.momentum
    wavefunction = vl * (fi[0] * fo[0] + fi[1] * fo[1]) + vr * (
        fi[2] * fo[2] + fi[3] * fo[3]
    )

    phi = ScalarWf(wavefunction=wavefunction, momentum=momentum, direction=1)

    return attach_scalar(phi, mass, width)


def current_fv_to_f(
    vertex: VertexFFV, mass: float, width: float, psi: DiracWf, polvec: VectorWf
) -> DiracWf:
    eps = polvec.wavefunction
    f = psi.wavefunction
    vl = vertex.left
    vr = vertex.right

    momentum = psi.momentum * psi.direction * polvec.momentum

    wavefunction = jax.lax.switch(
        (psi.direction + 1) // 2,
        # phi.(gl g[mu].PL + gr g[mu].PR) eps[mu]
        jnp.array(
            [
                vl * (f[3] * (eps[1] + im * eps[2]) + f[2] * (eps[0] + eps[3])),
                vl * (f[2] * (eps[1] - im * eps[2]) + f[3] * (eps[0] - eps[3])),
                vr * (-f[1] * (eps[1] + im * eps[2]) + f[0] * (eps[0] - eps[3])),
                vr * (-f[0] * (eps[1] - im * eps[2]) + f[1] * (eps[0] + eps[3])),
            ]
        ),
        # (gl g[mu].PL + gr g[mu].PR).psi eps[mu]
        jnp.array(
            [
                vr * (-f[3] * (eps[1] - im * eps[2]) + f[2] * (eps[0] - eps[3])),
                vr * (-f[2] * (eps[1] + im * eps[2]) + f[3] * (eps[0] + eps[3])),
                vl * (f[1] * (eps[1] - im * eps[2]) + f[0] * (eps[0] + eps[3])),
                vl * (f[0] * (eps[1] + im * eps[2]) + f[1] * (eps[0] - eps[3])),
            ]
        ),
    )

    psi = DiracWf(wavefunction=wavefunction, momentum=momentum, direction=psi.direction)
    return attach_dirac(psi, mass, width)


def current_ff_to_v(
    vertex: VertexFFV, mass: float, width: float, psi_out: DiracWf, psi_in: DiracWf
) -> VectorWf:
    fi = psi_in.wavefunction
    fo = psi_out.wavefunction
    vl = vertex.left
    vr = vertex.right

    momentum = psi_in.momentum - psi_out.momentum

    wavefunction = jnp.array(
        [
            vr * (fi[2] * fo[0] + fi[3] * fo[1]) + vl * (fi[0] * fo[2] + fi[1] * fo[3]),
            vr * (fi[3] * fo[0] + fi[2] * fo[1]) - vl * (fi[1] * fo[2] + fi[0] * fo[3]),
            im * vr * (fi[2] * fo[1] - fi[3] * fo[0])
            + im * vl * (fi[1] * fo[2] - fi[0] * fo[3]),
            vr * (fi[2] * fo[0] - fi[3] * fo[1]) + vl * (fi[1] * fo[3] - fi[0] * fo[2]),
        ]
    )

    return attach_vector(
        VectorWf(wavefunction=wavefunction, momentum=momentum, direction=1), mass, width
    )


# =============================================================================
# ---- Amplitudes -------------------------------------------------------------
# =============================================================================


def amplitude_ffv(
    vertex: VertexFFV, psi_out: DiracWf, psi_in: DiracWf, polvec: VectorWf
):
    # assert psi_out.direction == -1, "`psi_out` must be have flow out."
    # assert psi_in.direction == 1, "`psi_in` must be have flow in."

    fi = psi_in.wavefunction
    fo = psi_out.wavefunction
    eps = polvec.wavefunction
    vl = vertex.left
    vr = vertex.right
    im = 1.0j

    eps0p3 = eps[0] + eps[3]
    eps0m3 = eps[0] - eps[3]

    eps1p2 = eps[1] + im * eps[2]
    eps1m2 = eps[1] - im * eps[2]

    return vr * (
        fi[2] * (-fo[1] * eps1p2 + fo[0] * eps0m3)
        + fi[3] * (-fo[0] * eps1m2 + fo[1] * eps0p3)
    ) + vl * (
        fi[1] * (fo[2] * eps1m2 + fo[3] * eps0m3)
        + fi[0] * (fo[3] * eps1p2 + fo[2] * eps0p3)
    )


def amplitude_ffs(vertex: VertexFFS, psi_out: DiracWf, psi_in: DiracWf, phi: ScalarWf):
    # assert psi_out.direction == -1, "`psi_out` must be have flow out."
    # assert psi_in.direction == 1, "`psi_in` must be have flow in."

    fi = psi_in.wavefunction
    fo = psi_out.wavefunction
    vl = vertex.left
    vr = vertex.right
    return phi.wavefunction * (
        vl * (fi[0] * fo[0] + fi[1] * fo[1]) + vr * (fi[2] * fo[2] + fi[3] * fo[3])
    )



"""Module for computing currents and amplitudes."""

from .ffs import amplitude_ffs, current_ff_to_s, current_fs_to_f
from .ffv import amplitude_ffv, current_ff_to_v, current_fv_to_f
from .propagators import (attach_dirac, attach_scalar, attach_vector,
                          complex_mass_sqr, propagator_den)

__all__ = [
    "complex_mass_sqr",
    "propagator_den",
    "attach_scalar",
    "attach_vector",
    "attach_dirac",
    "current_ff_to_v",
    "current_fv_to_f",
    "amplitude_ffv",
    "current_ff_to_s",
    "current_fs_to_f",
    "amplitude_ffs",
]



"""Module for computing currents and amplitudes for two fermions and a scalar.
"""


import numpy as np

# from .dirac import WeylS0, WeylS1, WeylS2, WeylS3
from helax.numpy.wavefunctions import DiracWf, ScalarWf
from helax.vertices import VertexFFS

from .propagators import attach_dirac, attach_scalar

IM = 1.0j


# =============================================================================
# ---- Currents ---------------------------------------------------------------
# =============================================================================


def current_fs_to_f(
    vertex: VertexFFS, mass: float, width: float, psi: DiracWf, phi: ScalarWf
) -> DiracWf:
    """Fuse a fermion and scalar into an off-shell fermion.

    Parameters
    ----------
    vertex : VertexFFS
        Feynman rule for the F-F-S vertex.
    mass : float
        Mass of the resulting fermion.
    width : float
        Width of the resulting fermion
    psi : DiracWf
        Fermion to fuse with scalar.
    phi : ScalarWf
        Scalar to fuse with fermion.

    Returns
    -------
    chi: DiracWf
        Resulting fermion.
    """
    phi_wf = phi.wavefunction
    fi = psi.wavefunction
    vl = vertex.left
    vr = vertex.right

    momentum = psi.momentum * psi.direction + phi.momentum

    wavefunction = np.array(
        [
            vl * phi_wf * fi[0],
            vl * phi_wf * fi[1],
            vr * phi_wf * fi[2],
            vr * phi_wf * fi[3],
        ]
    )

    psi = DiracWf(wavefunction=wavefunction, momentum=momentum, direction=psi.direction)
    return attach_dirac(psi, mass, width)


def current_ff_to_s(
    vertex: VertexFFS, mass: float, width: float, psi_out: DiracWf, psi_in: DiracWf
) -> ScalarWf:
    """Fuse two fermions into a scalar.

    Parameters
    ----------
    vertex : VertexFFS
        Feynman rule for the F-F-S vertex.
    mass : float
        Mass of the resulting scalar.
    width : float
        Width of the resulting scalar.
    psi_out : DiracWf
        Flow-out fermion.
    psi_in : DiracWf
        Flow-in fermion.

    Returns
    -------
    phi: ScalarWf
        Resulting scalar wavefuction.
    """
    wf_in = psi_in.wavefunction
    wf_out = psi_out.wavefunction
    vl = vertex.left
    vr = vertex.right

    momentum = psi_in.momentum - psi_out.momentum
    wavefunction = vl * (wf_in[0] * wf_out[0] + wf_in[1] * wf_out[1]) + vr * (
        wf_in[2] * wf_out[2] + wf_in[3] * wf_out[3]
    )

    phi = ScalarWf(wavefunction=wavefunction, momentum=momentum, direction=1)

    return attach_scalar(phi, mass, width)


# =============================================================================
# ---- Amplitudes -------------------------------------------------------------
# =============================================================================


def amplitude_ffs(vertex: VertexFFS, psi_out: DiracWf, psi_in: DiracWf, phi: ScalarWf):
    """Compute the scattering amplitude for a Fermion-Fermion-Scalar vertex.

    Parameters
    ----------
    vertex : VertexFFS
        Feynman rule for the F-F-S vertex.
    psi_out : DiracWf
        Flow-out fermion wavefunction.
    psi_in : DiracWf
        Flow-in fermion wavefunction.
    phi : ScalarWf
        Scalar wavefunction.

    Returns
    -------
    amp: complex
        Scattering amplitude.
    """
    assert psi_out.direction == -1, "`psi_out` must be have flow out."
    assert psi_in.direction == 1, "`psi_in` must be have flow in."

    wf_in = psi_in.wavefunction
    wf_out = psi_out.wavefunction
    vl = vertex.left
    vr = vertex.right
    return phi.wavefunction * (
        vl * (wf_in[0] * wf_out[0] + wf_in[1] * wf_out[1])
        + vr * (wf_in[2] * wf_out[2] + wf_in[3] * wf_out[3])
    )



from .dirac import (DiracWf, charge_conjugate, spinor_u, spinor_ubar, spinor_v,
                    spinor_vbar)
from .scalar import ScalarWf, scalar_wf
from .vector import VectorWf, vector_wf

# from .weyl import (WeylType, WeylWf, weyl_lower_index, weyl_raise_index,
#                    weyl_x, weyl_xd, weyl_y, weyl_yd)

__all__ = [
    "DiracWf",
    "charge_conjugate",
    "spinor_u",
    "spinor_ubar",
    "spinor_v",
    "spinor_vbar",
    "ScalarWf",
    "scalar_wf",
    "VectorWf",
    "vector_wf",
    # "WeylType",
    # "WeylWf",
    # "weyl_lower_index",
    # "weyl_raise_index",
    # "weyl_x",
    # "weyl_xd",
    # "weyl_y",
    # "weyl_yd",
]



import unittest

import numpy as np
import pytest

from helax.numpy.dirac import WeylS0, WeylS1, WeylS2, WeylS3

from .weyl import (weyl_lower_index, weyl_raise_index, weyl_x, weyl_xd, weyl_y,
                   weyl_yd)


def _sigma_momentum(momentum, bar=False):
    sgn = -1 if not bar else 1
    return (
        momentum[0] * WeylS0
        + sgn * momentum[1] * WeylS1
        + sgn * momentum[2] * WeylS2
        + sgn * momentum[3] * WeylS3
    )


class TestWeylCompleteness(unittest.TestCase):
    def setUp(self) -> None:
        self.mass = 4.0
        self.momenta = np.transpose(
            np.array(
                [
                    [5.0, 0.0, 0.0, 3.0],
                    [5.0, 0.0, 0.0, -3.0],
                ]
            )
        )

    def test_completeness_weyl_x(self):
        # shapes: (num_spins, 2, num_momenta)
        wf_x = np.array(
            [weyl_x(self.momenta, self.mass, s).wavefunction for s in (-1, 1)]
        )
        wf_xd = np.array(
            [weyl_xd(self.momenta, self.mass, s).wavefunction for s in (-1, 1)]
        )

        for i in range(self.momenta.shape[-1]):
            x = wf_x[..., i]
            xd = wf_xd[..., i]
            spin_sum = np.einsum("ij,ik", x, xd)
            expect = _sigma_momentum(self.momenta[..., i])
            print()
            print(spin_sum)
            print(expect)
            print()
            # self.assertLess(np.max(np.abs(spin_sum - expect)), 1e-10)

    def test_completeness_weyl_y(self):
        # shapes: (num_spins, 2, num_momenta)
        wf_y = np.array(
            [weyl_y(self.momenta, self.mass, s).wavefunction for s in (-1, 1)]
        )
        wf_yd = np.array(
            [weyl_yd(self.momenta, self.mass, s).wavefunction for s in (-1, 1)]
        )

        for i in range(self.momenta.shape[-1]):
            y = wf_y[..., i]
            yd = wf_yd[..., i]
            spin_sum = np.einsum("ij,ik", yd, y)
            expect = _sigma_momentum(self.momenta[..., i], True)
            print()
            print(spin_sum)
            print(expect)
            print()
            # self.assertLess(np.max(np.abs(spin_sum - expect)), 1e-10)
